# coding=utf-8
#
# Copyright 2023 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
pen_handling.py

Classes for managing AxiDraw pen vertical motion and status, plus keeping track
of overall XYZ pen position.

Part of the AxiDraw driver for Inkscape
https://github.com/evil-mad/AxiDraw

The classes defined by this module are:

* PenPosition: Data storage class to hold XYZ pen position

* PenHandler: Main class for managing pen lifting, lowering, and status

* PenHeight: Manage pen-down height settings and keep timing up to date

* PenLiftTiming: Class to calculate and store pen lift timing settings

* PenStatus: Data storage class for pen lift status variables

"""

import time

from plotink import (
    ebb_motion,
    ebb_serial,  # https://github.com/evil-mad/plotink
)

# from ink_extensions import inkex


class PenPosition:
    """PenPosition: Class to store XYZ position of pen"""

    def __init__(self):
        self.xpos = 0  # X coordinate
        self.ypos = 0  # Y coordinate
        self.z_up = None  # Initialize as None: state unknown.

    def reset(self):
        """Reset XYZ positions to default."""
        self.xpos = 0
        self.ypos = 0
        self.z_up = None

    def reset_z(self):
        """Reset Z position only."""
        self.z_up = None


class PenHeight:
    """
    PenHeight: Class to manage pen-down height settings.
    Calculate timing for transiting between pen-up and pen-down states.
    """

    def __init__(self):
        self.pen_pos_down = None  # Initial values must be set by update().
        self.use_temp_pen_height = False  # Boolean set true while using temporary value
        self.narrow_band = False  # If true, use narrow band servo configuration.
        self.times = PenLiftTiming()

    def update(self, ad_ref):
        """
        Set initial/default values of options, after __init__.
        Call this function after changing option values to update pen height settings.
        """
        if not self.use_temp_pen_height:
            self.pen_pos_down = ad_ref.options.pen_pos_down
        self.times.update(ad_ref, self.narrow_band, self.pen_pos_down)

    def set_temp_height(self, ad_ref, temp_height):
        """
        Begin using temporary pen height position. Return True if the position has changed.
        """
        self.use_temp_pen_height = True
        if self.pen_pos_down == temp_height:
            return False
        self.pen_pos_down = temp_height

        self.times.update(ad_ref, self.narrow_band, temp_height)
        return True

    def end_temp_height(self, ad_ref):
        """
        End using temporary pen height position. Return True if the position has changed.
        """
        self.use_temp_pen_height = False
        if self.pen_pos_down == ad_ref.options.pen_pos_down:
            return False
        self.pen_pos_down = ad_ref.options.pen_pos_down
        self.times.update(ad_ref, self.narrow_band, self.pen_pos_down)
        return True


class PenLiftTiming:  # pylint: disable=too-few-public-methods
    """
    PenTiming: Class to calculate and store time required for pen to lift and lower
    """

    def __init__(self):
        self.raise_time = None
        self.lower_time = None

    def update(self, ad_ref, narrow_band, pen_down_pos):
        """
        Compute travel time needed for raising and lowering the pen.

        Call this function after changing option values to update pen timing settings.

        Servo travel time is estimated as the 4th power average (a smooth blend between):
          (A) Servo transit time for fast servo sweeps (t = slope * v_dist + min) and
          (B) Sweep time for slow sweeps (t = v_dist * full_scale_sweep_time / sweep_rate)
        """
        v_dist = abs(float(ad_ref.options.pen_pos_up - pen_down_pos))

        if narrow_band:
            servo_move_slope = ad_ref.params.nb_servo_move_slope
            servo_move_min = ad_ref.params.nb_servo_move_min
            servo_sweep_time = ad_ref.params.nb_servo_sweep_time
        else:
            servo_move_slope = ad_ref.params.servo_move_slope
            servo_move_min = ad_ref.params.servo_move_min
            servo_sweep_time = ad_ref.params.servo_sweep_time

        # Raising time:
        v_time = int(
            (
                (servo_move_slope * v_dist + servo_move_min) ** 4
                + (servo_sweep_time * v_dist / ad_ref.options.pen_rate_raise) ** 4
            )
            ** 0.25
        )
        if v_dist < 0.9:  # If up and down positions are equal, no initial delay
            v_time = 0

        v_time += ad_ref.options.pen_delay_up
        v_time = max(0, v_time)  # Do not allow negative total delay time
        self.raise_time = v_time

        # Lowering time:
        v_time = int(
            (
                (servo_move_slope * v_dist + servo_move_min) ** 4
                + (servo_sweep_time * v_dist / ad_ref.options.pen_rate_lower) ** 4
            )
            ** 0.25
        )
        if v_dist < 0.9:  # If up and down positions are equal, no initial delay
            v_time = 0
        v_time += ad_ref.options.pen_delay_down
        v_time = max(0, v_time)  # Do not allow negative total delay time
        self.lower_time = v_time


class PenStatus:
    """
    PenTiming: Data storage class for pen lift status variables

    preview_pen_state: pen state for preview rendering. 0: down, 1: up, -1: changed
    lifts: Counter; keeps track of the number of times the pen is lifted
    config: List of last [pen_pos_up, pen_pos_down, narrow_band]
    """

    def __init__(self):
        self.preview_pen_state = -1  # Will be moved to preview.py in the future
        self.lifts = 0
        self.config = [-1, -1, False]  # [pen_pos_up, pen_pos_down, narrow_band]

    def reset(self):
        """Clear preview pen state and lift count; Resetting them for a new plot."""
        self.preview_pen_state = -1  # Will be moved to preview.py in the future
        self.lifts = 0

    def report(self, ad_ref, message_fun):
        """report: Print pen lift statistics"""
        if not (ad_ref.options.report_time and ad_ref.params.report_lifts):
            return
        message_fun(f'Number of pen lifts: {self.lifts}\n')


class PenHandler:
    """
    PenHandler: Main class for managing pen lifting, lowering, and status,
    plus keeping track of XYZ pen position.
    """

    def __init__(self):
        self.heights = PenHeight()
        self.status = PenStatus()
        self.phys = PenPosition()  # Physical XYZ pen position
        self.turtle = PenPosition()  # turtle XYZ pen position, for interactive control

    def reset(self):
        """
        Reset certain defaults for a new plot:
        Clear pen height and lift count; clear temporary pen height flag.
        These are the defaults that can be set even before options are set.
        """
        self.status.reset()
        self.heights.use_temp_pen_height = False

    def pen_raise(self, ad_ref):
        """Raise the pen"""

        self.status.preview_pen_state = -1  # For preview rendering use

        # Skip if physical pen is already up:
        if self.phys.z_up:
            return

        self.status.lifts += 1

        v_time = self.heights.times.raise_time
        if self.heights.narrow_band:
            servo_pin = ad_ref.params.nb_servo_pin
        else:
            servo_pin = ad_ref.params.servo_pin

        if ad_ref.options.preview:
            ad_ref.preview.v_chart.rest(ad_ref, v_time)
        else:
            ebb_motion.sendPenUp(ad_ref.plot_status.port, v_time, servo_pin, False)
            if (v_time > 50) and (ad_ref.options.mode not in ['manual', 'align', 'cycle']):
                time.sleep(float(v_time - 30) / 1000.0)  # pause before issuing next command
            if ad_ref.params.use_b3_out:  # I/O Pin B3 output: low
                ebb_motion.PBOutValue(ad_ref.plot_status.port, 3, 0, False)
        self.phys.z_up = True

    def pen_lower(self, ad_ref):
        """Lower the pen"""

        self.status.preview_pen_state = -1  # For preview rendering use

        if self.phys.z_up is not None:
            if not self.phys.z_up:
                return  # skip if pen is state is _known_ and is down

        # Skip if stopped:
        if ad_ref.plot_status.stopped:
            return

        v_time = self.heights.times.lower_time

        if self.heights.narrow_band:
            servo_pin = ad_ref.params.nb_servo_pin
        else:
            servo_pin = ad_ref.params.servo_pin

        if ad_ref.options.preview:
            ad_ref.preview.v_chart.rest(ad_ref, v_time)
        else:
            ebb_motion.sendPenDown(ad_ref.plot_status.port, v_time, servo_pin, False)
            if (v_time > 50) and (ad_ref.options.mode not in ['manual', 'align', 'cycle']):
                time.sleep(float(v_time - 30) / 1000.0)  # pause before issuing next command
            if ad_ref.params.use_b3_out:  # I/O Pin B3 output: high
                ebb_motion.PBOutValue(ad_ref.plot_status.port, 3, 1, False)
        self.phys.z_up = False

    def cycle(self, ad_ref):
        """
        Toggle the pen down and then up, with a 1/2 second delay.
        Call only after servo_init(), which lowers the pen when initializing.
        This function should only be used as a setup utility.
        """
        ebb_motion.doTimedPause(ad_ref.plot_status.port, 500)
        self.pen_raise(ad_ref)

    def set_temp_height(self, ad_ref, temp_height):
        """Begin using temporary pen height position"""
        if self.heights.set_temp_height(ad_ref, temp_height):
            self.servo_init(ad_ref)

    def end_temp_height(self, ad_ref):
        """End use of temporary pen height position"""
        if self.heights.end_temp_height(ad_ref):
            self.servo_init(ad_ref)

    def find_pen_state(self, ad_ref):
        """
        Determine if physical pen state is initialized, and if so, if it is up or down.
        If this program has already determined it to be up or down, self.phys.z_up
            will be True or False, not None.
        If the hardware position has been set the EBBLV will store a code representing
            the pen-up position and whether that's a narrow-band position or not.
            Check to see if that has been set, and if so, use it to set self.phys.z_up
        If (and only if) z_up was None, but now we trust it to be reliable, then also set the
            config value to reflect the current settings. If it was not None (if the servo
            init routine was already run), leave the config settings alone to detect changes.
        Add special cases to handle setup functions that initially lower the pen.
        Return "layer code" and "servo config list"
        """

        if self.heights.narrow_band:
            layer_code = 71 + ad_ref.options.pen_pos_up // 2  # Possible range 71-121.
        else:
            layer_code = 1 + ad_ref.options.pen_pos_up // 2  # Possible range 1 - 51.

        config = [ad_ref.options.pen_pos_up, self.heights.pen_pos_down, self.heights.narrow_band]

        if self.phys.z_up is not None:  # Pen status _has_ already been set in software.
            return layer_code, config

        ebblv = ebb_motion.queryEBBLV(ad_ref.plot_status.port, False)
        ebb_pen_up = ebb_motion.QueryPenUp(ad_ref.plot_status.port, False)

        # Special case: The pen should go *down* when first initialized
        if (
            (ad_ref.options.mode == 'manual' and ad_ref.options.manual_cmd == 'lower_pen')
            or (ad_ref.options.mode == 'toggle' and ebb_pen_up)
            or ad_ref.options.mode == 'cycle'
        ):
            self.phys.z_up = False  # Flag to initialize state "with pen down."
            self.status.config[1] = -1  # Flag to require sending new pen-lower command.
            return layer_code, config

        if ad_ref.options.mode == 'toggle' or (
            ad_ref.options.mode == 'manual' and ad_ref.options.manual_cmd == 'raise_pen'
        ):
            self.phys.z_up = True  # Flag to initialize state "with pen up."
            self.status.config[0] = -1  # Flag to require sending new pen-raise command.
            return layer_code, config

        if bool(ebblv):  # If ebblv is not 0 or None...
            if int(ebblv) == layer_code:
                self.phys.z_up = ebb_pen_up
                if self.phys.z_up is True:  # Set config to skip initial pen raising:
                    self.status.config = [
                        ad_ref.options.pen_pos_up,
                        self.heights.pen_pos_down,
                        self.heights.narrow_band,
                    ]
        return layer_code, config

    def servo_init(self, ad_ref):
        """
        Utility function, used for setting or updating the various parameters that control
        the pen-lift servo motor. Function replaces servo_setup_wrapper() and servo_setup()
        in prior versions.

        Actions:
        1. If current pen up/down state is not yet known, query EBB to see if it has
            been initialized, or whether it is still in its default power-on state.
        2. Send EBB commands to set servo positions, lifting/lowering speeds,
            PWM channel count, timeout, and standard or narrow-band servo output.
        3. Put the servo into a known state, if necessary, by sending a pen-lift or
            lower command.

        Methods:
        When the EBB is reset, it goes to its default "pen up" position. QueryPenUp
        will tell us that the in the pen-up state. However, its actual position is the
        default, not the pen-up position that we've requested.

        To fix this, we could manually command the pen to either the pen-up or pen-down
        position. HOWEVER, that may take as much as five seconds in the very slowest
        speeds, and we want to skip that delay if the pen is already in the right place,
        for example if we're plotting after raising the pen, or plotting twice in a row.

        Solution: Use an otherwise unused EBB firmware variable (EBBLV), which is set to
        zero upon reset. If we set that value to be nonzero, and later find that it's still
        nonzero, we know that the servo position has been set (at least once) since reset.

        QueryPenUp _does not_ confirm that the pen is at the *requested* pen-up position, or
        that it was set to use the correct output pin. We encode the approximate pen-up height
        and narrow-band setting, and store that in EBBLV. At software start up, we read out EBBLV,
        to verify that the current PEN UP position is correct, and that we can skip the initial
        pen-raising movement if (1) the software does not yet know what the pen state is,
        but (2) the AxiDraw confirms that the pen height has been set, the pen is up, and that
        the pen-up height already matches what we were going to set it to.

        For changes after initialization, we use a list that] encodes the pen-up, pen-down and
        narrow-band settings. If (e.g.,) the pen-down height has changed while the pen is down,
        (e.g., Python interactive API update(), move to the new position.
        """

        if ad_ref.options.penlift == 3:
            self.heights.narrow_band = True
            servo_max = ad_ref.params.nb_servo_max
            servo_min = ad_ref.params.nb_servo_min
            servo_sweep_time = ad_ref.params.nb_servo_sweep_time
            pwm_period = 0.03  # Units are "ms / 100", since pen_rate_raise is a %.
            servo_pin = ad_ref.params.nb_servo_pin
        else:
            self.heights.narrow_band = False
            servo_max = ad_ref.params.servo_max
            servo_min = ad_ref.params.servo_min
            servo_sweep_time = ad_ref.params.servo_sweep_time
            pwm_period = 0.24  # 24 ms: 8 channels at 3 ms each (divided by 100 as above)
            servo_pin = ad_ref.params.servo_pin

        self.heights.update(ad_ref)  # Ensure heights and transit times are known

        if ad_ref.options.preview:
            self.phys.z_up = True
        if ad_ref.options.preview or ad_ref.plot_status.port is None:
            return

        layer_code, new_config = self.find_pen_state(ad_ref)

        servo_range = servo_max - servo_min
        servo_slope = float(servo_range) / 100.0

        ebb_motion.setPenUpPos(
            ad_ref.plot_status.port, int(round(servo_min + servo_slope * ad_ref.options.pen_pos_up)), False
        )
        ebb_motion.setPenDownPos(
            ad_ref.plot_status.port, int(round(servo_min + servo_slope * self.heights.pen_pos_down)), False
        )

        servo_rate_scale = float(servo_range) * pwm_period / servo_sweep_time

        ebb_motion.setPenUpRate(
            ad_ref.plot_status.port, int(round(servo_rate_scale * ad_ref.options.pen_rate_raise)), False
        )
        ebb_motion.setPenDownRate(
            ad_ref.plot_status.port, int(round(servo_rate_scale * ad_ref.options.pen_rate_lower)), False
        )

        if ad_ref.params.use_b3_out:  # Configure I/O Pin B3 for use
            ebb_motion.PBOutConfig(ad_ref.plot_status.port, 3, 0, False)  # output, low

        if self.phys.z_up is not None and (self.status.config == new_config):
            return  # Servo is initialized. No changes to height or I/O pin are needed.

        # If pen-up or pen-down position has changed while up/down, make that change.
        if (self.phys.z_up is False) and (
            (self.status.config[1] != new_config[1]) or (self.status.config[2] != new_config[2])
        ):
            v_time = self.heights.times.lower_time
            ebb_motion.sendPenDown(ad_ref.plot_status.port, v_time, servo_pin, False)
            if ad_ref.params.use_b3_out:  # I/O Pin B3 output: high
                ebb_motion.PBOutValue(ad_ref.plot_status.port, 3, 1, False)
        if (self.phys.z_up is not False) and (
            (self.status.config[0] != new_config[0]) or (self.status.config[2] != new_config[2])
        ):
            v_time = self.heights.times.raise_time
            ebb_motion.sendPenUp(ad_ref.plot_status.port, v_time, servo_pin, False)
            if ad_ref.params.use_b3_out:  # I/O Pin B3 output: low
                ebb_motion.PBOutValue(ad_ref.plot_status.port, 3, 0, False)
            self.phys.z_up = True

        self.status.config = new_config
        ebb_motion.setEBBLV(ad_ref.plot_status.port, layer_code, False)

        if self.heights.narrow_band:
            ebb_serial.command(ad_ref.plot_status.port, 'SC,8,1\r')  # 1 channel of servo PWM
        else:
            ebb_serial.command(ad_ref.plot_status.port, 'SC,8,8\r')  # 8 channels of servo PWM
            ebb_motion.servo_timeout(
                ad_ref.plot_status.port, ad_ref.params.servo_timeout, None, False
            )  # Power timeout is only applicable to standard servo



#
# Copyright 2023 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
dripfeed.py

Manage, or simulate, the process of feeding individual motion segments to the AxiDraw

Part of the AxiDraw driver for Inkscape
https://github.com/evil-mad/AxiDraw

Requires Python 3.7 or newer.

"""

import logging
import time

from plotink import ebb_motion


def feed(ad_ref, move_list):
    """
    Feed individual motion actions to the AxiDraw during a plot or preview.
    Take care of housekeeping while doing so, including:
        Checking for pause inputs
        Skipping physical moves while in preview mode
        Skipping physical moves while resuming plots
        Updating previews
        Updating node counts
        Updating CLI progress bar
        Keeping track of total distance traveled, pen-up and pen-down
        Sleeping during long moves
        Reporting errors to the user
    Inputs: AxiDraw reference object, list of movement commands
    """

    if move_list is None:
        return

    spew_dripfeed_debug_data = False  # Set True to get entirely too much debugging data

    drip_logger = logging.getLogger('.'.join([__name__, 'dripfeed']))
    if spew_dripfeed_debug_data:
        drip_logger.setLevel(logging.DEBUG)  # by default level is INFO

    # drip_logger.debug('\ndripfeed.feed()\n')
    # drip_logger.debug('move_list:\n' + str(move_list)) # Can print full move list

    for move in move_list:
        ad_ref.pause_check()

        if ad_ref.plot_status.stopped:
            ad_ref.plot_status.copies_to_plot = 0
            return
        if ad_ref.pen.phys.xpos is None:
            return  # Physical location is not well-defined; stop here.

        if move[0] == 'lower':
            ad_ref.pen.pen_lower(ad_ref)
            continue

        if move[0] == 'raise':
            ad_ref.pen.pen_raise(ad_ref)
            continue

        if move[0] == 'SM':
            feed_sm(ad_ref, move, drip_logger)
            continue


def feed_sm(ad_ref, move, drip_logger):
    """
    Manage the process of sending a single "SM" move command to the AxiDraw,
        and simulate doing so when in preview mode.
    Take care of housekeeping while doing so, including:
        Skipping physical moves while in preview mode
        Updating previews
        Updating progress bar (CLI)
        Keeping track of total distance traveled, pen-up and pen-down
        Sleeping during long moves
        Reporting errors to the user
    """

    # drip_logger.debug('\ndripfeed.feed_SM()\n')

    # 'SM' move is formatted as:
    # ['SM', (move_steps2, move_steps1, move_time), seg_data]
    # where seg_data begins with:
    #   * final x position, float
    #   * final y position, float
    #   * final pen_up state, boolean
    #   * travel distance (inch)

    move_steps2 = move[1][0]
    move_steps1 = move[1][1]
    move_time = move[1][2]
    f_new_x = move[2][0]
    f_new_y = move[2][1]
    move_dist = move[2][3]

    if ad_ref.options.preview:
        ad_ref.plot_status.stats.pt_estimate += move_time
        # log_sm_for_preview(ad_ref, move)

        ad_ref.preview.log_sm_move(ad_ref, move)

    else:
        ebb_motion.doXYMove(ad_ref.plot_status.port, move_steps2, move_steps1, move_time, False)

        if move_time > 50:  # Sleep before issuing next command
            if ad_ref.options.mode != 'manual':
                time.sleep(float(move_time - 30) / 1000.0)
    # drip_logger.debug('XY move: (%s, %s), in %s ms', move_steps1, move_steps2, move_time)
    # drip_logger.debug('fNew(X,Y): (%.5f, %.5f)', f_new_x, f_new_y)

    ad_ref.plot_status.stats.add_dist(ad_ref.pen.phys.z_up, move_dist)  # Distance; inches
    ad_ref.plot_status.progress.update_auto(ad_ref.plot_status.stats)

    ad_ref.pen.phys.xpos = f_new_x  # Update current position
    ad_ref.pen.phys.ypos = f_new_y


def page_layer_delay(ad_ref, between_pages=True, delay_ms=None):
    """
    Execute page delay or layer delay, monitoring for pause signals.
    Set between_pages=True for page delays, false for layer delays.
    delay_ms is only used for layer delays.
    """

    if ad_ref.plot_status.stopped:
        return  # No delay if stopped.

    if between_pages:
        if ad_ref.plot_status.copies_to_plot == 0:
            return  # No delay after last copy, for page delays.
        ad_ref.plot_status.delay_between_copies = True  # Set flag: Delaying between copies
        delay_ms = ad_ref.options.page_delay * 1000

    if not delay_ms:  # If delay time is 0 or None, exit.
        return
    if delay_ms >= 1000:  # Only launch progress bar for at least 1 s of delay time
        ad_ref.plot_status.progress.launch_sub(ad_ref, delay_ms, page=between_pages)

    # Number of rest intervals:
    sleep_interval = 100  # Time period to sleep, ms. Default: 100
    time_remaining = delay_ms

    while time_remaining > 0:
        if ad_ref.plot_status.stopped:
            break  # Exit loop if stopped.
        if time_remaining < 150:  # If less than 150 ms left to delay,
            sleep_interval = time_remaining  # do it all at once.

        if between_pages:
            ad_ref.plot_status.stats.page_delays += sleep_interval
        else:
            ad_ref.plot_status.stats.layer_delays += sleep_interval

        if ad_ref.options.preview:
            ad_ref.plot_status.stats.pt_estimate += sleep_interval
        else:
            time.sleep(sleep_interval / 1000)  # Use short intervals for responsiveness
            ad_ref.plot_status.progress.update_sub_rel(sleep_interval)  # update progress bar
            ad_ref.pause_check()  # Detect button press while between plots
        time_remaining -= sleep_interval
    ad_ref.plot_status.progress.close_sub()
    delay_between_copies = False



#
# Copyright 2023 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
digest_svg.py

Routines for digesting an input SVG file into a simplified SVG ("Plob")
formats for easier processing

Part of the AxiDraw driver for Inkscape
https://github.com/evil-mad/AxiDraw

Requires Python 3.7 or newer.
"""

import logging
from math import sqrt

from ink_extensions import cubicsuperpath, inkex, simplepath, simplestyle, simpletransform
from ink_extensions_utils import message
from lxml import etree

# https://github.com/evil-mad/plotink
from plotink import plot_utils

from . import path_objects

logger = logging.getLogger(__name__)


class DigestSVG:  # pylint: disable=pointless-string-statement
    """
    Main class for parsing SVG document and "digesting" it into a
    path_objects.DocDigest object, a heavily simplified representation of the
    SVG contents.
    """

    logging_attrs = {'default_handler': message.UserMessageHandler()}
    spew_debugdata = False

    def __init__(self, default_logging=True):
        # Create instance variables

        if default_logging:
            logger.setLevel(logging.INFO)
            logger.addHandler(self.logging_attrs['default_handler'])
        if self.spew_debugdata:
            logger.setLevel(logging.DEBUG)  # by default level is INFO

        self.use_tag_nest_level = 0
        self.current_layer = None
        self.current_layer_name = ''
        self.next_id = 0

        self.doc_digest = path_objects.DocDigest()

        # Variables that will be populated in process_svg():
        self.bezier_tolerance = 0
        self.layer_selection = -2  # All layers; Matches default from plot_status.py

        self.doc_width_100 = 0
        self.doc_height_100 = 0
        self.diagonal_100 = 0

    def process_svg(self, node_list, warnings, digest_params, mat_current=None):
        """
        Wrapper around routine to recursively traverse an SVG document.

        This calls the recursive routine and handles building the digest
        structure around it, as well as reporting any necessary errors.

        Inputs:
        node_list, an lxml etree representing an SVG document
        warnings, a plot_warnings.PlotWarnings object
        digest_params, a tuple with additional parameters
        mat_current, a transformation matrix

        From the SVG, build and return a path_objects.DocDigest object; a
        python object representing a simplified SVG document. Once flattened,
        the digest can be converted to a Plob file if desired.

        The Plob (Plot Object) format is a highly-restricted subset of SVG.
            Only layers are allowed in the SVG root. Only polylines are allowed
            in layers. There are no other objects, nor object transformations
            or styles.
        """

        [
            self.doc_digest.width,
            self.doc_digest.height,
            scale_x,
            scale_y,
            self.layer_selection,
            self.bezier_tolerance,
        ] = digest_params

        # Store document information in doc_digest
        self.doc_digest.viewbox = f'0 0 {self.doc_digest.width:f} {self.doc_digest.height:f}'

        self.doc_width_100 = self.doc_digest.width / scale_x  # Width of a "100% width" object
        self.doc_height_100 = self.doc_digest.height / scale_y  # height of a "100% height" object
        self.diagonal_100 = sqrt((self.doc_width_100) ** 2 + (self.doc_height_100) ** 2) / sqrt(2)

        docname = node_list.get('{http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd}docname')
        if docname:  # Previously: inkex.addNS('docname', 'sodipodi'),
            self.doc_digest.name = docname

        root_layer = path_objects.LayerItem()
        # Flag to identify layers containing objects from root
        root_layer.name = '__digest-root__'
        root_layer.item_id = str(self.next_id)
        self.next_id += 1
        self.doc_digest.layers.append(root_layer)

        self.current_layer = root_layer  # Layer that graphical elements should be added to
        self.current_layer_name = root_layer.name

        self.traverse(node_list, None, warnings, mat_current)
        return self.doc_digest

    def traverse(self, node_list, parent_style, warnings, mat_current):
        """
        Recursively traverse the SVG file and process all of the paths. Keep
        track of the composite transformation applied to each path.

        Inputs:
        node_list, an lxml etree representing an SVG document
        parent_style, dict from inherit_style
        warnings, a plot_warnings.PlotWarnings object
        mat_current, a transformation matrix

        This function handles path, group, line, rect, polyline, polygon,
        circle, ellipse  and use (clone) elements. Notable elements not handled
        include text. Unhandled elements should be converted to paths in
        Inkscape or another vector graphics editor.
        """

        if mat_current is None:
            mat_current = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]

        for node in node_list:
            node_visibility = node.get('visibility')
            element_style = simplestyle.parseStyle(node.get('style'))

            # Presentation attributes, which have lower precedence than the style attribute:
            if 'fill' not in element_style:  # If the style has not been set...
                element_style['fill'] = node.get('fill')
            if 'stroke' not in element_style:  # If the style has not been set...
                element_style['stroke'] = node.get('stroke')
            if 'fill-rule' not in element_style:  # If the style has not been set...
                element_style['fill-rule'] = node.get('fill-rule')
            # Since these are added to the style dictionary, a potential problem is that
            # these are now treated on equal footing to CSS styling information.

            style_dict = inherit_style(parent_style, element_style, node_visibility)

            if style_dict['display'] == 'none':
                continue  # Do not plot this object or its children
            if node.get('display') == 'none':  # Possible SVG attribute as well
                continue  # Do not plot this object or its children

            # Apply the current matrix transform to this node's transform
            trans = node.get('transform')
            if trans is None:
                mat_new = mat_current
            else:
                mat_new = simpletransform.composeTransform(mat_current, simpletransform.parseTransform(trans))

            if node.tag in ('{http://www.w3.org/2000/svg}g', 'g'):
                old_layer_name = self.current_layer_name
                if (
                    old_layer_name == '__digest-root__'
                    and node.get('{http://www.inkscape.org/namespaces/inkscape}groupmode') == 'layer'
                ):
                    # Ensure that sublayers are treated like regular groups only

                    str_layer_name = node.get('{http://www.inkscape.org/namespaces/inkscape}label')
                    if str_layer_name is None:
                        str_layer_name = f'Auto-Layer {self.next_id}'

                    new_layer = path_objects.LayerItem()
                    new_layer.name = str_layer_name
                    new_layer.parse_name()

                    if new_layer.props.skip:
                        continue  # Skip Documentation layer and its contents
                    if self.layer_selection >= 0:  # Plotting in layers mode
                        if new_layer.props.number is None:
                            continue
                        if self.layer_selection != new_layer.props.number:
                            continue  # Skip this layer and its contents

                    new_layer.item_id = str(self.next_id)
                    self.next_id += 1
                    self.doc_digest.layers.append(new_layer)
                    self.current_layer = new_layer
                    self.current_layer_name = str(str_layer_name)

                    self.traverse(node, style_dict, warnings, mat_new)

                    # After parsing a layer, add a new "root layer" for any objects
                    # that may appear in root before the next layer:

                    new_layer = path_objects.LayerItem()
                    new_layer.name = '__digest-root__'  # Label this as a "root" layer
                    new_layer.item_id = str(self.next_id)
                    self.next_id += 1

                    self.doc_digest.layers.append(new_layer)
                    self.current_layer = new_layer
                    self.current_layer_name = new_layer.name
                else:  # Regular group or sublayer that we treat as a group.
                    self.traverse(node, style_dict, warnings, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}symbol', 'symbol'):
                # A symbol is much like a group, except that it should only
                #       be rendered when called within a "use" tag.
                if self.use_tag_nest_level > 0:
                    self.traverse(node, style_dict, warnings, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}a', 'a'):
                # An 'a' is much like a group, in that it is a generic container element.
                self.traverse(node, style_dict, warnings, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}switch', 'switch'):
                # A 'switch' is much like a group, in that it is a generic container element.
                # We are not presently evaluating conditions on switch elements, but parsing
                # their contents to the extent possible.
                self.traverse(node, style_dict, warnings, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}use', 'use'):
                """
                A <use> element refers to another SVG element via an xlink:href="#blah"
                attribute.  We will handle the element by doing an XPath search through
                the document, looking for the element with the matching id="blah"
                attribute.  We then recursively process that element after applying
                any necessary (x,y) translation.

                Notes:
                 1. We ignore the height and g attributes as they do not apply to
                    path-like elements, and
                 2. Even if the use element has visibility="hidden", SVG still calls
                    for processing the referenced element.  The referenced element is
                    hidden only if its visibility is "inherit" or "hidden".
                 3. We may be able to unlink clones using the code in pathmodifier.py
                """

                refid = node.get('{http://www.w3.org/1999/xlink}href')
                if refid is not None:
                    # [1:] to ignore leading '#' in reference
                    path = f'//*[@id="{refid[1:]}"]'
                    refnode = node.xpath(path)
                    if refnode is not None:
                        x_val = float(node.get('x', '0'))
                        y_val = float(node.get('y', '0'))
                        # Note: the transform has already been applied
                        if x_val != 0 or y_val != 0:
                            mat_new2 = simpletransform.composeTransform(
                                mat_new, simpletransform.parseTransform(f'translate({x_val:.6E},{y_val:.6E})')
                            )
                        else:
                            mat_new2 = mat_new
                        self.use_tag_nest_level += 1  # Keep track of nested "use" elements.
                        self.traverse(refnode, style_dict, warnings, mat_new2)
                        self.use_tag_nest_level -= 1
                continue

            # End container elements; begin graphical elements.

            if self.layer_selection >= 0:
                if self.current_layer_name == '__digest-root__':
                    continue  # Do not print root elements if layer_selection >= 0

            if style_dict['visibility'] in ('hidden', 'collapse'):
                # Not visible; Do not plot. (This comes after the container tags;
                #   visible children of hidden elements can still plot.)
                continue

            if node.tag == '{http://www.w3.org/2000/svg}path':
                path_d = node.get('d')
                self.digest_path(path_d, style_dict, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}rect', 'rect'):
                """
                Create a path with the outline of the rectangle
                Manually transform  <rect x="X" y="Y" width="W" height="H"/>
                    into            <path d="MX,Y lW,0 l0,H l-W,0 z"/>
                Draw three sides of the rectangle explicitly and the fourth implicitly
                https://www.w3.org/TR/SVG11/shapes.html#RectElement
                """

                x = plot_utils.unitsToUserUnits(node.get('x', '0'), self.doc_width_100)
                y = plot_utils.unitsToUserUnits(node.get('y', '0'), self.doc_height_100)

                r_x, width = [
                    plot_utils.unitsToUserUnits(node.get(attr), self.doc_width_100) for attr in ['rx', 'width']
                ]
                r_y, height = [
                    plot_utils.unitsToUserUnits(node.get(attr), self.doc_height_100) for attr in ['ry', 'height']
                ]

                def calc_r_attr(attr, other_attr, twice_maximum):
                    value = attr if attr is not None else other_attr if other_attr is not None else 0
                    return min(value, twice_maximum * 0.5)

                r_x = calc_r_attr(r_x, r_y, width)
                r_y = calc_r_attr(r_y, r_x, height)

                instr = []
                if (r_x > 0) or (r_y > 0):
                    instr.append(['M ', [x + r_x, y]])
                    instr.append([' L ', [x + width - r_x, y]])
                    instr.append([' A ', [r_x, r_y, 0, 0, 1, x + width, y + r_y]])
                    instr.append([' L ', [x + width, y + height - r_y]])
                    instr.append([' A ', [r_x, r_y, 0, 0, 1, x + width - r_x, y + height]])
                    instr.append([' L ', [x + r_x, y + height]])
                    instr.append([' A ', [r_x, r_y, 0, 0, 1, x, y + height - r_y]])
                    instr.append([' L ', [x, y + r_y]])
                    instr.append([' A ', [r_x, r_y, 0, 0, 1, x + r_x, y]])
                else:
                    instr.append(['M ', [x, y]])
                    instr.append([' L ', [x + width, y]])
                    instr.append([' L ', [x + width, y + height]])
                    instr.append([' L ', [x, y + height]])
                    instr.append([' L ', [x, y]])

                self.digest_path(simplepath.formatPath(instr), style_dict, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}line', 'line'):
                """
                Convert an SVG line object  <line x1="X1" y1="Y1" x2="X2" y2="Y2/>
                to an SVG path object:      <path d="MX1,Y1 LX2,Y2"/>
                """
                x_1, x_2 = [
                    plot_utils.unitsToUserUnits(node.get(attr, '0'), self.doc_width_100) for attr in ['x1', 'x2']
                ]
                y_1, y_2 = [
                    plot_utils.unitsToUserUnits(node.get(attr, '0'), self.doc_height_100) for attr in ['y1', 'y2']
                ]

                path_a = []
                path_a.append(['M ', [x_1, y_1]])
                path_a.append([' L ', [x_2, y_2]])
                self.digest_path(simplepath.formatPath(path_a), style_dict, mat_new)
                continue

            if node.tag in (
                '{http://www.w3.org/2000/svg}polyline',
                'polyline',
                '{http://www.w3.org/2000/svg}polygon',
                'polygon',
            ):
                """
                Convert
                 <polyline points="x1,y1 x2,y2 x3,y3 [...]"/>
                OR
                 <polyline points="x1 y1 x2 y2 x3 y3 [...]"/>
                OR
                 <polygon points="x1,y1 x2,y2 x3,y3 [...]"/>
                OR
                 <polygon points="x1 y1 x2 y2 x3 y3 [...]"/>
                to
                  <path d="Mx1,y1 Lx2,y2 Lx3,y3 [...]"/> (with a closing Z on polygons)
                Ignore polylines with no points, or polylines with only a single point.
                """

                pl = node.get('points', '').strip()
                if pl == '':
                    continue
                pa = pl.replace(',', ' ').split()  # replace comma with space before splitting
                if not pa:
                    continue
                path_length = len(pa)
                if path_length < 4:  # Minimum of x1,y1 x2,y2 required.
                    continue
                path_d = 'M ' + pa[0] + ' ' + pa[1]
                i = 2
                while i < (path_length - 1):
                    path_d += ' L ' + pa[i] + ' ' + pa[i + 1]
                    i += 2

                if node.tag in ('{http://www.w3.org/2000/svg}polygon', 'polygon'):
                    path_d += ' Z'
                self.digest_path(path_d, style_dict, mat_new)
                continue

            if node.tag in (
                '{http://www.w3.org/2000/svg}ellipse',
                'ellipse',
                '{http://www.w3.org/2000/svg}circle',
                'circle',
            ):
                """
                Convert circles and ellipses to paths as two 180 degree arcs.
                In general (an ellipse), we convert
                  <ellipse rx="RX" ry="RY" cx="X" cy="Y"/>
                to
                  <path d="MX1,CY A RX,RY 0 1 0 X2,CY A RX,RY 0 1 0 X1,CY"/>
                where
                  X1 = CX - RX
                  X2 = CX + RX
                Ellipses or circles with a radius attribute of 0 are ignored
                """

                if node.tag in ('{http://www.w3.org/2000/svg}circle', 'circle'):
                    r_x = plot_utils.unitsToUserUnits(node.get('r', '0'), self.diagonal_100)
                    r_y = r_x
                else:
                    r_x, r_y = [
                        plot_utils.unitsToUserUnits(node.get(attr, '0'), self.diagonal_100) for attr in ['rx', 'ry']
                    ]
                if r_x == 0 or r_y == 0:
                    continue

                c_x = plot_utils.unitsToUserUnits(node.get('cx', '0'), self.doc_width_100)
                c_y = plot_utils.unitsToUserUnits(node.get('cy', '0'), self.doc_height_100)

                x_1 = c_x - r_x
                x_2 = c_x + r_x
                path_d = (
                    f'M {x_1:f},{c_y:f} '
                    + f'A {r_x:f},{r_y:f} '
                    + f'0 1 0 {x_2:f},{c_y:f} '
                    + f'A {r_x:f},{r_y:f} '
                    + f'0 1 0 {x_1:f},{c_y:f}'
                )
                self.digest_path(path_d, style_dict, mat_new)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}metadata', 'metadata'):
                self.doc_digest.metadata.update(dict(node.attrib))
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}plotdata', 'plotdata'):
                self.doc_digest.plotdata.update(dict(node.attrib))
                continue

            if node.tag in [
                '{http://www.w3.org/2000/svg}defs',
                'defs',
                'namedview',
                '{http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd}namedview',
                'eggbot',
                'WCB',
                'MergeData',
                '{http://www.w3.org/2000/svg}eggbot',
                '{http://www.w3.org/2000/svg}WCB',
                '{http://www.w3.org/2000/svg}MergeData',
                '{http://www.w3.org/2000/svg}title',
                'title',
                '{http://www.w3.org/2000/svg}desc',
                'desc',
                '{http://www.w3.org/2000/svg}pattern',
                'pattern',
                '{http://www.w3.org/2000/svg}radialGradient',
                'radialGradient',
                '{http://www.w3.org/2000/svg}linearGradient',
                'linearGradient',
                '{http://www.w3.org/2000/svg}style',
                'style',  #  external style sheet
                '{http://www.w3.org/2000/svg}cursor',
                'cursor',
                '{http://www.w3.org/2000/svg}font',
                'font',
                '{http://www.inkscape.org/namespaces/inkscape}templateinfo',
                '{http://www.w3.org/2000/svg}color-profile',
                'color-profile',
                '{http://www.w3.org/2000/svg}foreignObject',
                'foreignObject',
                etree.Comment,
            ]:
                continue

            if node.tag in (
                '{http://www.w3.org/2000/svg}text',
                'text',
                '{http://www.w3.org/2000/svg}flowRoot',
                'flowRoot',
            ):
                warnings.add_new('text', self.current_layer_name)
                continue

            if node.tag in ('{http://www.w3.org/2000/svg}image', 'image'):
                warnings.add_new('image', self.current_layer_name)
                continue
            if not isinstance(node.tag, str):
                # This is likely an XML processing instruction such as an XML
                # comment. lxml uses a function reference for such node tags
                # and as such the node tag is likely not a printable string.
                # Converting it to a printable string likely won't be useful.
                continue
            text = str(node.tag).split('}')
            warnings.add_new(str(text[-1]), self.current_layer_name)

    def digest_path(self, path_d, style_dict, mat_transform):
        """
        Parse the path while applying the matrix transformation mat_transform.
        - Input is the "d" string attribute from an SVG path.
        - Turn this path into a cubicsuperpath (list of beziers).
        - Subdivide the cubic path into a list, rendering to straight segments within tolerance.
        - Identify subpaths within each path
        - Build a path_objects.PathItem object and append it to the current layer
        """

        # logger.debug('digest_path()\n')
        # logger.debug('path d: ' + path_d)

        if path_d is None:
            return
        if path_d == '':
            return

        parsed_path = cubicsuperpath.CubicSuperPath(simplepath.parsePath(path_d))

        if len(parsed_path) == 0:  # path length is zero, will not be plotted
            return

        # Apply the transformation to each point
        apply_transform_to_path(mat_transform, parsed_path)

        subpaths = []

        # p is now a list of lists of cubic beziers [control pt1, control pt2, endpoint]
        # where the start-point is the last point in the previous segment.
        for subpath in parsed_path:  # for subpaths in the path:
            # Divide each path into a set of straight segments:
            plot_utils.subdivideCubicPath(subpath, self.bezier_tolerance)

            if len(subpath) < 2:
                continue  # At least two points required for a path
            # Pick out vertex location information from cubic bezier curve:
            subpaths.append([[vertex[1][0], vertex[1][1]] for vertex in subpath])

        if len(subpaths) == 0:
            return  # At least one sub-path required

        new_path = path_objects.PathItem()
        new_path.fill = style_dict['fill']
        new_path.stroke = style_dict['stroke']
        new_path.fill_rule = style_dict['fill-rule']
        new_path.item_id = str(self.next_id)
        self.next_id += 1

        new_path.subpaths = subpaths

        ok_to_fill = False
        for subpath in subpaths:
            if len(subpath) != 2:
                ok_to_fill = True  # As long as at least one path has more than two vertices
                break
        if not ok_to_fill:
            new_path.fill = None  # Strip fill, if path has only 2-vertex subpaths

        # Add new list of subpaths to the current "LayerItem" element:
        self.current_layer.paths.append(new_path)

        # logger.debug('End of digest_path()\n')


def apply_transform_to_path(mat, path):
    """
    A very slightly faster version of simpletransform.applyTransformToPath()
    Possibly move this function to plotink in the future.
    """
    [mt00, mt01, mt02], [mt10, mt11, mt12] = mat
    for comp in path:
        for ctl in comp:
            for point in ctl:  # apply transform to each point:
                pt_x = point[0]
                pt_y = point[1]
                point[0] = mt00 * pt_x + mt01 * pt_y + mt02
                point[1] = mt10 * pt_x + mt11 * pt_y + mt12


def inherit_style(parent_style, node_style, visibility):
    """
    Parse style dict of node for fill and stroke information only.
    Inherit style from parent, but supersede it when a local style is defined.
    Also handle precedence of SVG "visibility" attribute, separate from the style.
    Note that children of hidden parents may be plotted if they assert visibility.
    """
    default_style = {}
    default_style['fill'] = None
    default_style['stroke'] = None
    default_style['fill-rule'] = None
    default_style['visibility'] = 'visible'
    default_style['display'] = None  # A null value; not "display:none".

    if parent_style is None:  # Use default values when there is no parent
        parent_style = default_style

    # Use copy, not assignment, so that new_style represents an independent dict:
    new_style = parent_style.copy()

    if visibility:  # Update first, allowing it to be overruled by style attributes
        new_style['visibility'] = visibility

    if node_style is None:  # No additional new style information provided.
        return new_style

    for attrib in [
        'fill',
        'stroke',
        'fill-rule',
        'visibility',
        'display',
    ]:
        # Valid for "string" attributes that DO NOT have units that need scaling;
        # Do not extend this to other style attributes without accounting for that.
        value = node_style.get(attrib)  # Defaults to None, preventing KeyError
        if value:
            if value in ['inherit']:
                new_style[attrib] = parent_style[attrib]
            else:
                new_style[attrib] = value

    return new_style


def verify_plob(svg, model):
    """
    Check to see if the provided SVG is a valid plob that can be automatically converted
    to a plot digest object. Also check that the plob version and hardware model match.
    Styles are *presently allowed* in the plob, for the sake of verification. (They can
        increase the file size, but do not otherwise cause harm.)

    Returns True or False.

    We may wish to also check for an application name match in the
    future. At present, that check is not yet necessary.
    """

    data_node = None
    nodes = svg.xpath('//*[self::svg:plotdata|self::plotdata]', namespaces=inkex.NSS)
    if nodes:
        data_node = nodes[0]
    if data_node is not None:
        try:
            svg_model = data_node.get('model')
            svg_plob_version = data_node.get('plob_version')
        except TypeError:
            return False
    else:
        return False  # No plot data; Plob cannot be verified.
    if svg_model:
        if int(svg_model) != model:
            return False
    else:
        return False
    if svg_plob_version:
        if svg_plob_version != path_objects.PLOB_VERSION:
            return False
    else:
        return False

    # inkex.errormsg( "Passed plotdata checks") # Optional halfwaypoint check
    for node in svg:
        if node.tag in ['g', '{http://www.w3.org/2000/svg}g']:
            name_temp = node.get('{http://www.inkscape.org/namespaces/inkscape}label')
            if name_temp is None:
                return False  # All groups must be named
            if len(str(name_temp)) > 0:
                if str(name_temp)[0] == '%':
                    continue  # Skip Documentation layer and its contents
            if node.get('transform'):  # No transforms are allowed on plottable layers
                return False
            for subnode in node:
                if subnode.get('transform'):  # No transforms are allowed on objects
                    return False
                if subnode.tag in ['polyline', '{http://www.w3.org/2000/svg}polyline']:
                    continue
                return False
        elif node.tag in [
            '{http://www.w3.org/2000/svg}defs',
            'defs',
            'metadata',
            '{http://www.w3.org/2000/svg}metadata',
            '{http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd}namedview',
            'plotdata',
            '{http://www.w3.org/2000/svg}plotdata',
        ]:
            continue
        else:
            return False
    return True



#
# Copyright 2021 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
boundsclip.py

Functions for clipping

Part of the AxiDraw driver for Inkscape
https://github.com/evil-mad/AxiDraw
"""

from plotink import plot_utils


def clip_at_bounds(digest, phy_bounds, doc_bounds, warn_tol, doc_clip=True):
    """
    Step through subpaths in the digest, clipping them at plot
    boundaries, splitting them into additional subpaths if necessary.

    The plot boundary is defined by (0,0) and a maximum X and Y value
    which may be defined by the physical bounds, or both the physical
    bounds and the document bounds. Return a warning message if the
    requested travel exceeds the bounds by greater than the tolerance.

    Inputs:
    digest: A path_objects.DocDigest object
    phy_bounds: Physical bounds. A 4-element list:
        [[x_min, y_min],[x_max, y_max]], adjusted out for rounding tolerance
    doc_bounds: Document bounds. A 2-element list with (x_max, y_max)
    warn_tol: Distance that motion exceed limits without warning messages
    doc_clip: Boolean. If True, clip at both phy_bounds and doc_bounds

    Output: Warning text or None

    Flatten the document prior to clipping if it is not already flattened.

    This routine performs no rotation; digest orientation should match
    orientation for which the bounds are given.

    Notes on boundaries and warnings about clipping:

    Generate a warning only if the requested motion:
    (1) Exceeds the physical bounds by at least the value of warn_tol, AND
    (2) Is clipped by physical limits, rather than the page size, AND
    (3) Is clipped in the positive direction (not at X = 0 or Y = 0).

    No warning will be generated when travel is limited by a document
    size smaller than the travel, nor at the lower limits of travel.

    Both sets of bounds are typically set slightly past the nominal
    boundaries, by a value typically 1e-9. This ensures that clipping is not
    applied to values that are within float precision of the nominal bounds.
    For example, if a square has a side at Y position at nominal zero but
    numericly  -1.05E-14, we do not want to clip, and thereby omit, that side.
    """
    clip_warn_x = True  # Enable warnings about X clipping
    clip_warn_y = True  # Enable warnings about Y clipping

    out_of_bounds_flag = False  # No warning yet generated

    # Positive limits:
    [x_max, y_max] = phy_bounds[1]

    if doc_clip:
        [page_max_x, page_max_y] = doc_bounds
        if x_max >= page_max_x:
            clip_warn_x = False  # Limited by page size, not travel size
            x_max = page_max_x
        if y_max >= page_max_y:
            clip_warn_y = False  # Limited by page size, not travel size
            y_max = page_max_y

    clip_bounds = [phy_bounds[0], [x_max, y_max]]

    # Loose tolerance bounds for generating warning messages:
    x_max_warn = x_max + warn_tol
    y_max_warn = y_max + warn_tol

    digest.flatten()

    for layer in digest.layers:  # Each layer is a LayerItem object.
        for path in layer.paths:  # Each path is a PathItem object.
            input_subpath = path.subpaths[0]
            new_subpaths = []
            a_subpath = []
            first_point = True
            prev_in_bounds = False
            prev_vertex = []

            for vertex in input_subpath:
                [v_x, v_y] = vertex

                in_bounds = plot_utils.point_in_bounds(vertex, clip_bounds, 0)
                if not in_bounds:
                    # Only check if there's no warning issued yet
                    if not out_of_bounds_flag:
                        if clip_warn_x:
                            if v_x > x_max_warn:
                                out_of_bounds_flag = True
                        if clip_warn_y:
                            if v_y > y_max_warn:
                                out_of_bounds_flag = True
                """
                Clipping logic:

                Possible cases, for first vertex:
                (1) In bounds: Add the vertex to the path.
                (2) Not in bounds: Do not add the vertex.

                Possible cases, for subsequent vertices:
                (1) In bounds, as was previous: Add the vertex.
                  -> No segment between two in-bound points needs clipping.
                (2) In bounds, prev was not: Clip & start new path.
                (3) OOB, prev was in bounds: Clip & end the path.
                (4) OOB, as was previous: Segment _may_ clip corner.
                  -> Either add no points or start & end new path
                """

                if first_point:
                    if in_bounds:
                        a_subpath.append(vertex)
                else:
                    if in_bounds and prev_in_bounds:
                        a_subpath.append(vertex)
                    else:
                        segment = [prev_vertex, vertex]
                        accept, seg = plot_utils.clip_segment(segment, clip_bounds)
                        if accept:
                            if in_bounds and not prev_in_bounds:
                                if len(a_subpath) > 0:
                                    new_subpaths.append(a_subpath)
                                    a_subpath = []  # start new subpath
                                a_subpath.append([seg[0][0], seg[0][1]])
                                v_x = seg[1][0]
                                v_y = seg[1][1]
                                a_subpath.append([v_x, v_y])
                            if prev_in_bounds and not in_bounds:
                                v_x = seg[1][0]
                                v_y = seg[1][1]
                                a_subpath.append([v_x, v_y])
                                new_subpaths.append(a_subpath)  # Save subpath
                                a_subpath = []  # Start new subpath
                            if (not prev_in_bounds) and not in_bounds:
                                if len(a_subpath) > 0:
                                    new_subpaths.append(a_subpath)
                                    a_subpath = []  # start new subpath
                                a_subpath.append([seg[0][0], seg[0][1]])
                                v_x = seg[1][0]
                                v_y = seg[1][1]
                                a_subpath.append([v_x, v_y])
                                new_subpaths.append(a_subpath)  # Save subpath
                                a_subpath = []  # Start new subpath
                        else:
                            in_bounds = False
                first_point = False
                prev_vertex = vertex
                prev_in_bounds = in_bounds

            if len(a_subpath) > 0:
                new_subpaths.append(a_subpath)
            path.subpaths = new_subpaths
        layer.flatten()  # Re-flatten layer
    return out_of_bounds_flag



#
# SVG Path Ordering Extension
# This extension uses a simple TSP algorithm to order the paths so as
# to reduce plotting time by plotting nearby paths consecutively.
#
#
# While written from scratch, this is a derivative in spirit of the work by
# Matthew Beckler and Daniel C. Newman for the EggBot project.
#
# The MIT License (MIT)
#
# Copyright (c) 2021 Windell H. Oskay, Evil Mad Science LLC
# www.evilmadscientist.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math

from ink_extensions import inkex, simplestyle, simpletransform
from ink_extensions_utils import exit_status

# import sys
from lxml import etree
from plotink import plot_utils  # https://github.com/evil-mad/plotink  Requires version 0.

"""
TODOs:

* Apparent difference in execution time for portrait vs landscape document orientation.
  Seems to be related to the _change_

* Implement path functions

<param name="path_handling" _gui-text="Compound Paths" type="optiongroup">
<_option value=0>Leave as is</_option>
<_option value=1>Reorder subpaths</_option>
<_option value=2>Break apart</_option>
</param>

self.arg_parser.add_argument( "--path_handling",\
action="store", type=int, dest="path_handling",\
default=1,help="How compound paths are handled")


* Consider re-introducing GUI method for rendering:

<param indent="1" name="rendering" type=inkex.boolean_option _gui-text="Preview pen-up travel">
  false</param>


"""


class ReorderEffect(inkex.Effect):
    """
    Inkscape effect extension.
    Re-order the objects in the SVG document for faster plotting.
    Respect layers: Initialize a new dictionary of objects for each layer, and sort
        objects within that layer only
    Objects in root of document are treated as being on a _single_ layer, and will all
        be sorted.

    """

    def __init__(self):
        inkex.Effect.__init__(self)

        self.arg_parser.add_argument(
            '--reordering', action='store', type=int, dest='reordering', default=1, help='How groups are handled'
        )

        self.auto_rotate = True

    def effect(self):
        # Main entry point of the program

        self.svg_width = 0
        self.svg_height = 0
        self.air_total_default = 0
        self.air_total_sorted = 0
        self.printPortrait = False

        # Rendering is available for debug purposes. It only previews
        # pen-up movements that are reordered and typically does not
        # include all possible movement.

        self.preview_rendering = False
        self.layer_index = 0  # index for coloring layers

        self.svg = self.document.getroot()

        self.DocUnits = 'in'  # Default
        self.DocUnits = self.getDocumentUnit()

        self.unit_scaling = 1

        self.getDocProps()

        """
        Set up the document-wide transforms to handle SVG viewbox 
        """

        matCurrent = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]

        viewbox = self.svg.get('viewBox')

        vb = self.svg.get('viewBox')
        if vb:
            p_a_r = self.svg.get('preserveAspectRatio')
            sx, sy, ox, oy = plot_utils.vb_scale(vb, p_a_r, self.svg_width, self.svg_height)
        else:
            sx = 1.0 / float(plot_utils.PX_PER_INCH)  # Handle case of no viewbox
            sy = sx
            ox = 0.0
            oy = 0.0

        # Initial transform of document is based on viewbox, if present:
        matCurrent = simpletransform.parseTransform(
            'scale({0:.6E},{1:.6E}) translate({2:.6E},{3:.6E})'.format(sx, sy, ox, oy)
        )
        # Set up x_last, y_last, which keep track of last known pen position
        # The initial position is given by the expected initial pen position

        self.y_last = 0

        if self.printPortrait:
            self.x_last = self.svg_width
        else:
            self.x_last = 0

        parent_vis = 'visible'

        self.root_nodes = []

        if self.preview_rendering:
            # Remove old preview layers, if rendering is enabled
            for node in self.svg:
                if node.tag == inkex.addNS('g', 'svg') or node.tag == 'g':
                    if node.get(inkex.addNS('groupmode', 'inkscape')) == 'layer':
                        LayerName = node.get(inkex.addNS('label', 'inkscape'))
                        if LayerName == '% Preview':
                            self.svg.remove(node)

            preview_transform = simpletransform.parseTransform(
                'translate({2:.6E},{3:.6E}) scale({0:.6E},{1:.6E})'.format(1.0 / sx, 1.0 / sy, -ox, -oy)
            )
            path_attrs = {'transform': simpletransform.formatTransform(preview_transform)}
            self.preview_layer = etree.Element(inkex.addNS('g', 'svg'), path_attrs, nsmap=inkex.NSS)

            self.preview_layer.set(inkex.addNS('groupmode', 'inkscape'), 'layer')
            self.preview_layer.set(inkex.addNS('label', 'inkscape'), '% Preview')
            self.svg.append(self.preview_layer)

            # Preview stroke width: 1/1000 of page width or height, whichever is smaller
            if self.svg_width < self.svg_height:
                width_du = self.svg_width / 1000.0
            else:
                width_du = self.svg_height / 1000.0

            """
            Stroke-width is a css style element, and cannot accept scientific notation.
            
            Thus, in cases with large scaling (i.e., high values of 1/sx, 1/sy)
            resulting from the viewbox attribute of the SVG document, it may be necessary to use 
            a _very small_ stroke width, so that the stroke width displayed on the screen
            has a reasonable width after being displayed greatly magnified thanks to the viewbox.
            
            Use log10(the number) to determine the scale, and thus the precision needed.
            """

            log_ten = math.log10(width_du)
            if log_ten > 0:  # For width_du > 1
                width_string = '{0:.3f}'.format(width_du)
            else:
                prec = int(math.ceil(-log_ten) + 3)
                width_string = '{0:.{1}f}'.format(width_du, prec)

            self.p_style = {
                'stroke-width': width_string,
                'fill': 'none',
                'stroke-linejoin': 'round',
                'stroke-linecap': 'round',
            }

        self.svg = self.parse_svg(self.svg, matCurrent)

    def parse_svg(self, input_node, mat_current=None, parent_vis='visible'):
        """
        Input: An SVG node (usually) containing other nodes:
            The SVG root, a layer, sublayer, or other group.
        Output: The re-ordered node. The contents are reordered with the greedy
            algorithm, except:
            - Layers and sublayers are preserved. The contents of each are
                    re-ordered for faster plotting.
            - Groups are either preserved, broken apart, or re-ordered within
                    the group, depending on the value of group_mode.
        """

        coord_dict = {}
        # coord_dict maps a node ID to the following data:
        #    Is the node plottable, first coordinate pair, last coordinate pair.
        #    i.e., Node_id -> (Boolean: plottable, Xi, Yi, Xf, Yf)

        group_dict = {}
        # group_dict maps a node ID for a group to the contents of that group.
        # The contents may be a preserved nested group or a flat list, depending
        #  on the selected group handling mode. Example:
        # group_dict = {'id_1': <Element {http://www.w3.org/2000/svg}g at memory_location_1>,
        #               'id_2': <Element {http://www.w3.org/2000/svg}g at memory_location_2>

        nodes_to_delete = []

        counter = 0  # TODO: Replace this with better unique ID system

        # Account for input_node's transform and any transforms above it:
        if mat_current is None:
            mat_current = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]
        try:
            matNew = simpletransform.composeTransform(
                mat_current, simpletransform.parseTransform(input_node.get('transform'))
            )
        except AttributeError:
            matNew = mat_current

        for node in input_node:
            # Step through each object within the top-level input node

            if node.tag is etree.Comment:
                continue

            try:
                id = node.get('id')
            except AttributeError:
                id = self.uniqueId('1', True)
                node.set('id', id)
            if id == None:
                id = self.uniqueId('1', True)
                node.set('id', id)

            # First check for object visibility:
            skip_object = False

            # Check for "display:none" in the node's style attribute:
            style = simplestyle.parseStyle(node.get('style'))
            if 'display' in style.keys() and style['display'] == 'none':
                skip_object = True  # Plot neither this object nor its children

            # The node may have a display="none" attribute as well:
            if node.get('display') == 'none':
                skip_object = True  # Plot neither this object nor its children

            # Visibility attributes control whether a given object will plot.
            # Children of hidden (not visible) parents may be plotted if
            # they assert visibility.
            visibility = node.get('visibility', parent_vis)

            if 'visibility' in style.keys():
                visibility = style['visibility']  # Style may override attribute.

            if visibility == 'inherit':
                visibility = parent_vis

            if visibility != 'visible':
                skip_object = True  # Skip this object and its children

            # Next, check to see if this inner node is itself a group or layer:
            if node.tag == inkex.addNS('g', 'svg') or node.tag == 'g':
                # Use the user-given option to decide what to do with subgroups:
                subgroup_mode = self.options.reordering

                #                 Values of the parameter:
                #                 subgroup_mode=="1": Preserve groups
                #                 subgroup_mode=="2": Reorder within groups
                #                 subgroup_mode=="3": Break apart groups

                if node.get(inkex.addNS('groupmode', 'inkscape')) == 'layer':
                    # The node is a layer or sub-layer, not a regular group.
                    # Parse it separately, and re-order its contents.

                    subgroup_mode = 2  # Always sort within each layer.
                    self.layer_index += 1

                    layer_name = node.get(inkex.addNS('label', 'inkscape'))

                    layer_name = str(layer_name).lstrip()

                    if layer_name:
                        if layer_name[0] == '%':  # First character is '%'; This
                            skip_object = True  # is a documentation layer; skip plotting.
                            self.layer_index -= 1  # Set this back to previous value.

                if skip_object:
                    # Do not re-order hidden groups or layers.
                    subgroup_mode = 1  # Preserve this group

                if subgroup_mode == 3:
                    # Break apart this non-layer subgroup and add it to
                    # the set of things to be re-ordered.

                    nodes_to_delete.append(node)
                    nodes_inside_group = self.group2NodeDict(node)

                    for a_node in nodes_inside_group:
                        try:
                            id = a_node.get('id')
                        except AttributeError:
                            id = self.uniqueId('1', True)
                            a_node.set('id', id)
                        if id == None:
                            id = self.uniqueId('1', True)
                            a_node.set('id', id)

                        # Use getFirstPoint and getLastPoint on each object:
                        start_plottable, first_point = self.getFirstPoint(a_node, matNew)
                        end_plottable, last_point = self.getLastPoint(a_node, matNew)

                        coord_dict[id] = (
                            start_plottable and end_plottable,
                            first_point[0],
                            first_point[1],
                            last_point[0],
                            last_point[1],
                        )
                        # Entry in group_dict is this node
                        group_dict[id] = a_node

                elif subgroup_mode == 2:
                    # Reorder a layer or subgroup with a recursive call.

                    node = self.parse_svg(node, matNew, visibility)

                    # Capture the first and last x,y coordinates of the optimized node
                    start_plottable, first_point = self.group_first_pt(node, matNew)
                    end_plottable, last_point = self.group_last_pt(node, matNew)

                    # Then add this optimized node to the coord_dict
                    coord_dict[id] = (
                        start_plottable and end_plottable,
                        first_point[0],
                        first_point[1],
                        last_point[0],
                        last_point[1],
                    )
                    # Entry in group_dict is this node
                    group_dict[id] = node

                else:  # (subgroup_mode == 1)
                    # Preserve the group, but find its first and last point so
                    #    that it can be re-ordered with respect to other items

                    if skip_object:
                        start_plottable = False
                        end_plottable = False
                        first_point = [(-1.0), (-1.0)]
                        last_point = [(-1.0), (-1.0)]
                    else:
                        start_plottable, first_point = self.group_first_pt(node, matNew)
                        end_plottable, last_point = self.group_last_pt(node, matNew)

                    coord_dict[id] = (
                        start_plottable and end_plottable,
                        first_point[0],
                        first_point[1],
                        last_point[0],
                        last_point[1],
                    )
                    # Entry in group_dict is this node
                    group_dict[id] = node

            else:  # Handle objects that are not groups
                if skip_object:
                    start_plottable = False
                    end_plottable = False
                    first_point = [(-1.0), (-1.0)]
                    last_point = [(-1.0), (-1.0)]
                else:
                    start_plottable, first_point = self.getFirstPoint(node, matNew)
                    end_plottable, last_point = self.getLastPoint(node, matNew)

                coord_dict[id] = (
                    start_plottable and end_plottable,
                    first_point[0],
                    first_point[1],
                    last_point[0],
                    last_point[1],
                )
                group_dict[id] = node  # Entry in group_dict is this node

        # Perform the re-ordering:
        ordered_element_list = self.ReorderNodeList(coord_dict, group_dict)

        # Once a better order for the svg elements has been determined,
        # All there is do to is to reintroduce the nodes to the parent in the correct order
        for elt in ordered_element_list:
            # Creates identical node at the correct location according to ordered_element_list
            input_node.append(elt)
        # Once program is finished parsing through
        for element_to_remove in nodes_to_delete:
            try:
                input_node.remove(element_to_remove)
            except ValueError:
                inkex.errormsg(str(element_to_remove.get('id')) + ' is not a member of ' + str(input_node.get('id')))

        return input_node

    def break_apart_path(self, path):
        """
        An SVG path may contain multiple distinct portions, that are normally separated
        by pen-up movements.

        This function takes the path data string from an SVG path, parses it, and returns
        a dictionary of independent path data strings, each of which represents a single
        pen-down movement. It is equivalent to the Inkscape function Path > Break Apart

        Input: path data string, representing a single SVG path
        Output: Dictionary of (separated) path data strings

        """
        MaxLength = len(path)
        ix = 0
        move_to_location = []
        path_dictionary = {}
        path_list = []
        path_number = 1

        # Search for M or m location
        while ix < MaxLength:
            if path[ix] == 'm' or path[ix] == 'M':
                move_to_location.append(ix)
            ix = ix + 1
        # Iterate through every M or m location in our list of move to instructions
        # Slice the path string according to path beginning and ends as indicated by the
        # location of these instructions

        for counter, m in enumerate(move_to_location):
            if m == move_to_location[-1]:
                # last entry
                path_list.append(path[m:MaxLength].rstrip())
            else:
                path_list.append(path[m : move_to_location[counter + 1]].rstrip())

        for counter, current_path in enumerate(path_list):
            # Enumerate over every entry in the path looking for relative m commands
            if current_path[0] == 'm' and counter > 0:
                # If path contains relative m command, the best case is when the last command
                # was a Z or z. In this case, all relative operations are performed relative to
                # initial x, y coordinates of the previous path

                # TODO this is not defines...here just placeholder!!!
                matNew = None
                if path_list[counter - 1][-1].upper() == 'Z':
                    current_path_x, current_path_y, index = self.getFirstPoint(current_path, matNew)
                    prev_path_x, prev_path_y, ignore = self.getFirstPoint(path_list[counter - 1])
                    adapted_x = current_path_x + prev_path_x
                    adapted_y = current_path_y + prev_path_y
                    # Now we can replace the path data with an Absolute Move to instruction
                    # HOWEVER, we need to adapt all the data until we reach a different command in the case of a repeating
                    path_list[counter] = 'm ' + str(adapted_x) + ',' + str(adapted_y) + ' ' + current_path[index:]

                # If there is no z or absolute commands, we need to parse the entire path
                else:
                    # scan path for absolute coordinates. If present, begin parsing from their index
                    # instead of the beginning
                    prev_path = path_list[counter - 1]
                    prev_path_length = len(prev_path)
                    jx = 0
                    x_val, y_val = 0, 0
                    # Check one char at a time
                    # until we have the moveTo Command
                    last_command = ''
                    is_absolute_command = False
                    repeated_command = False
                    # name of command
                    # how many parameters we need to skip
                    accepted_commands = {'M': 0, 'L': 0, 'H': 0, 'V': 0, 'C': 4, 'S': 2, 'Q': 2, 'T': 0, 'A': 5}

                    # If there is an absolute command which specifies a new initial point
                    # then we can save time by setting our index directly to its location in the path data
                    # See if an accepted_command is present in the path data. If it is present further in the
                    # string than any command found before, then set the pointer to that location
                    # if a command is not found, find() will return a -1. jx is initialized to 0, so if no matches
                    # are found, the program will parse from the beginning to the end of the path

                    for keys in 'MLCSQTA':  # TODO: Compare to last_point; see if we can clean up this part
                        if prev_path.find(keys) > jx:
                            jx = prev_path.find(keys)

                    while jx < prev_path_length:
                        temp_x_val = ''
                        temp_y_val = ''
                        num_of_params_to_skip = 0

                        # SVG Path commands can be repeated
                        if prev_path[jx].isdigit() and last_command:
                            repeated_command = True
                        else:
                            repeated_command = False

                        if (prev_path[jx].isalpha() and prev_path[jx].upper() in accepted_commands) or repeated_command:
                            if repeated_command:
                                # is_relative_command is saved from last iteration of the loop
                                current_command = last_command
                            else:
                                # If the character is accepted, we must parse until reach the x y coordinates
                                is_absolute_command = prev_path[jx].isupper()
                                current_command = prev_path[jx].upper()

                            # Each command has a certain number of parameters we must pass before we reach the
                            # information we care about. We will parse until we know that we have reached them

                            # Get to start of next number
                            # We will know we have reached a number if the current character is a +/- sign
                            # or current character is a digit
                            while jx < prev_path_length:
                                if prev_path[jx] in '+-' or prev_path[jx].isdigit():
                                    break
                                jx = jx + 1

                            # We need to parse past the unused parameters in our command
                            # The number of parameters to parse past is dependent on the command and stored
                            # as the value of accepted_command
                            # Spaces and commas are used to deliniate parameters
                            while jx < prev_path_length and num_of_params_to_skip < accepted_commands[current_command]:
                                if prev_path[jx].isspace() or prev_path[jx] == ',':
                                    num_of_params_to_skip = num_of_params_to_skip + 1
                                jx = jx + 1

                            # Now, we are in front of the x character

                            if current_command.upper() == 'V':
                                temp_x_val = 0

                            if current_command.upper() == 'H':
                                temp_y_val = 0

                            # Parse until next character is a digit or +/- character
                            while jx < prev_path_length and current_command.upper() != 'V':
                                if prev_path[jx] in '+-' or prev_path[jx].isdigit():
                                    break
                                jx = jx + 1

                            # Save each next character until we reach a space
                            while (
                                jx < prev_path_length
                                and current_command.upper() != 'V'
                                and not (prev_path[jx].isspace() or prev_path[jx] == ',')
                            ):
                                temp_x_val = temp_x_val + prev_path[jx]
                                jx = jx + 1

                            # Then we know we have completely parsed the x character

                            # Now we are in front of the y character

                            # Parse until next character is a digit or +/- character
                            while jx < prev_path_length and current_command.upper() != 'H':
                                if prev_path[jx] in '+-' or prev_path[jx].isdigit():
                                    break
                                jx = jx + 1

                            ## Save each next character until we reach a space
                            while (
                                jx < prev_path_length
                                and current_command.upper() != 'H'
                                and not (prev_path[jx].isspace() or prev_path[jx] == ',')
                            ):
                                temp_y_val = temp_y_val + prev_path[jx]
                                jx = jx + 1

                            # Then we know we have completely parsed the y character

                            if is_absolute_command:
                                if current_command == 'H':
                                    # Absolute commands create new x,y position
                                    try:
                                        x_val = float(temp_x_val)
                                    except ValueError:
                                        pass
                                elif current_command == 'V':
                                    # Absolute commands create new x,y position
                                    try:
                                        y_val = float(temp_y_val)
                                    except ValueError:
                                        pass
                                else:
                                    # Absolute commands create new x,y position
                                    try:
                                        x_val = float(temp_x_val)
                                        y_val = float(temp_y_val)
                                    except ValueError:
                                        pass
                            else:
                                if current_command == 'h':
                                    # Absolute commands create new x,y position
                                    try:
                                        x_val = x_val + float(temp_x_val)
                                    except ValueError:
                                        pass
                                elif current_command == 'V':
                                    # Absolute commands create new x,y position
                                    try:
                                        y_val = y_val + float(temp_y_val)
                                    except ValueError:
                                        pass
                                else:
                                    # Absolute commands create new x,y position
                                    try:
                                        x_val = x_val + float(temp_x_val)
                                        y_val = y_val + float(temp_y_val)
                                    except ValueError:
                                        pass
                            last_command = current_command
                        jx = jx + 1
                    x, y, index = self.getFirstPoint(current_path, None)
                    path_list[counter] = 'm ' + str(x_val + x) + ',' + str(y_val + y) + ' ' + current_path[index:]

        for counter, path in enumerate(path_list):
            path_dictionary['ad_path' + str(counter)] = path

        return path_dictionary

    def getFirstPoint(self, node, matCurrent):
        """
        Input: (non-group) node and parent transformation matrix
        Output: Boolean value to indicate if the svg element is plottable and
            two floats stored in a list representing the x and y coordinates we plot first
        """

        # first apply the current matrix transform to this node's transform
        matNew = simpletransform.composeTransform(matCurrent, simpletransform.parseTransform(node.get('transform')))

        point = [float(-1), float(-1)]
        try:
            if node.tag == inkex.addNS('path', 'svg'):
                pathdata = node.get('d')

                point = plot_utils.pathdata_first_point(pathdata)
                if point:
                    simpletransform.applyTransformToPoint(matNew, point)
                    return True, point
                else:
                    return False, [float(-1), float(-1)]

            if node.tag == inkex.addNS('rect', 'svg') or node.tag == 'rect':
                """
                The x,y coordinates for a rect are included in their specific attributes
                If there is a transform, we need translate the x & y coordinates to their
                correct location via applyTransformToPoint.
                """

                point[0] = float(node.get('x'))
                point[1] = float(node.get('y'))

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('line', 'svg') or node.tag == 'line':
                """
                The x1 and y1 attributes are where we will start to draw
                So, get them, apply the transform matrix, and return the point
                """

                point[0] = float(node.get('x1'))
                point[1] = float(node.get('y1'))

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            elif node.tag in [inkex.addNS('polyline', 'svg'), 'polyline', inkex.addNS('polygon', 'svg'), 'polygon']:
                """
                Polyline and polygon have the same first point.

                We need to extract x1 and y1 from these:
                <polygon points="x1,y1 x2,y2 x3,y3 [...]"/>
                We accomplish this with Python string strip
                and split methods. Then apply transforms
                """
                pl = node.get('points', '').strip()

                if pl == '':
                    return False, point

                pa = pl.replace(',', ' ').split()  # replace comma with space before splitting

                if not pa:
                    return False, point
                pathLength = len(pa)
                if pathLength < 4:  # Minimum of x1,y1 x2,y2 required.
                    return False, point

                d = 'M ' + pa[0] + ' ' + pa[1]
                i = 2
                while i < (pathLength - 1):
                    d += ' L ' + pa[i] + ' ' + pa[i + 1]
                    i += 2

                point = plot_utils.pathdata_first_point(d)
                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('ellipse', 'svg') or node.tag == 'ellipse':
                cx = float(node.get('cx', '0'))
                cy = float(node.get('cy', '0'))
                rx = float(node.get('rx', '0'))

                point[0] = cx - rx
                point[1] = cy

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('circle', 'svg') or node.tag == 'circle':
                cx = float(node.get('cx', '0'))
                cy = float(node.get('cy', '0'))
                r = float(node.get('r', '0'))
                point[0] = cx - r
                point[1] = cy

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('symbol', 'svg') or node.tag == 'symbol':
                # A symbol is much like a group, except that
                # it's an invisible object.
                return False, point  # Skip this element.

            if node.tag == inkex.addNS('use', 'svg') or node.tag == 'use':
                """
                A <use> element refers to another SVG element via an xlink:href="#blah"
                attribute.  We will handle the element by doing an XPath search through
                the document, looking for the element with the matching id="blah"
                attribute.  We then recursively process that element after applying
                any necessary (x,y) translation.
                
                Notes:
                 1. We ignore the height and g attributes as they do not apply to
                    path-like elements, and
                 2. Even if the use element has visibility="hidden", SVG still calls
                    for processing the referenced element.  The referenced element is
                    hidden only if its visibility is "inherit" or "hidden".
                 3. We may be able to unlink clones using the code in pathmodifier.py
                """

                refid = node.get(inkex.addNS('href', 'xlink'))

                if refid is not None:
                    # [1:] to ignore leading '#' in reference
                    path = '//*[@id="{0}"]'.format(refid[1:])
                    refnode = node.xpath(path)
                    if refnode is not None:
                        x = float(node.get('x', '0'))
                        y = float(node.get('y', '0'))

                        # Note: the transform has already been applied
                        if x != 0 or y != 0:
                            mat_new2 = simpletransform.composeTransform(
                                matNew, simpletransform.parseTransform('translate({0:f},{1:f})'.format(x, y))
                            )
                        else:
                            mat_new2 = matNew
                        # Note that the referenced object may be a 'symbol`,
                        # which acts like a group, or it may be a simple
                        # object.

                        if len(refnode) > 0:
                            plottable, the_point = self.group_first_pt(refnode[0], mat_new2)
                        else:
                            plottable, the_point = self.group_first_pt(refnode, mat_new2)

                        return plottable, the_point
        except:
            pass

        # Svg Object is not a plottable element
        # In this case, return False to indicate a non-plottable element
        # and a default point

        return False, point

    def getLastPoint(self, node, matCurrent):
        """
        Input: XML tree node and transformation matrix
        Output: Boolean value to indicate if the svg element is plottable or not and
                two floats stored in a list representing the x and y coordinates we plot last
        """

        # first apply the current matrix transform to this node's transform
        matNew = simpletransform.composeTransform(matCurrent, simpletransform.parseTransform(node.get('transform')))

        # If we return a negative value, we know that this function did not work
        point = [float(-1), float(-1)]
        try:
            if node.tag == inkex.addNS('path', 'svg'):
                path = node.get('d')

                point = plot_utils.pathdata_last_point(path)
                if point:
                    simpletransform.applyTransformToPoint(matNew, point)
                    return True, point
                else:
                    return False, [float(-1), float(-1)]
            if node.tag == inkex.addNS('rect', 'svg') or node.tag == 'rect':
                """
                The x,y coordinates for a rect are included in their specific attributes
                If there is a transform, we need translate the x & y coordinates to their
                correct location via applyTransformToPoint.
                """

                point[0] = float(node.get('x'))
                point[1] = float(node.get('y'))

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point  # Same start and end points

            if node.tag == inkex.addNS('line', 'svg') or node.tag == 'line':
                """
                The x2 and y2 attributes are where we will end our drawing
                So, get them, apply the transform matrix, and return the point
                """

                point[0] = float(node.get('x2'))
                point[1] = float(node.get('y2'))

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('polyline', 'svg') or node.tag == 'polyline':
                pl = node.get('points', '').strip()

                if pl == '':
                    return False, point

                pa = pl.replace(',', ' ').split()
                if not pa:
                    return False, point
                pathLength = len(pa)
                if pathLength < 4:  # Minimum of x1,y1 x2,y2 required.
                    return False, point

                d = 'M ' + pa[0] + ' ' + pa[1]
                i = 2
                while i < (pathLength - 1):
                    d += ' L ' + pa[i] + ' ' + pa[i + 1]
                    i += 2

                endpoint = plot_utils.pathdata_last_point(d)
                simpletransform.applyTransformToPoint(matNew, endpoint)

                return True, endpoint

            elif node.tag in [inkex.addNS('polygon', 'svg'), 'polygon']:
                """
                Polygon has same first and last point.
                
                Repeat function to get first point of polyline:
                """
                pl = node.get('points', '').strip()

                if pl == '':
                    return False, point

                pa = pl.replace(',', ' ').split()  # replace comma with space before splitting

                if not pa:
                    return False, point
                pathLength = len(pa)
                if pathLength < 4:  # Minimum of x1,y1 x2,y2 required.
                    return False, point

                d = 'M ' + pa[0] + ' ' + pa[1]
                i = 2
                while i < (pathLength - 1):
                    d += ' L ' + pa[i] + ' ' + pa[i + 1]
                    i += 2

                point = plot_utils.pathdata_first_point(d)
                simpletransform.applyTransformToPoint(matNew, point)

            if node.tag == inkex.addNS('ellipse', 'svg') or node.tag == 'ellipse':
                cx = float(node.get('cx', '0'))
                cy = float(node.get('cy', '0'))
                rx = float(node.get('rx', '0'))

                point[0] = cx - rx
                point[1] = cy

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('circle', 'svg') or node.tag == 'circle':
                cx = float(node.get('cx', '0'))
                cy = float(node.get('cy', '0'))
                r = float(node.get('r', '0'))
                point[0] = cx - r
                point[1] = cy

                simpletransform.applyTransformToPoint(matNew, point)

                return True, point

            if node.tag == inkex.addNS('symbol', 'svg') or node.tag == 'symbol':
                # A symbol is much like a group, except that it should only be
                # rendered when called within a "use" tag.
                return False, point  # Skip this element.

            if node.tag == inkex.addNS('use', 'svg') or node.tag == 'use':
                """
                A <use> element refers to another SVG element via an xlink:href="#blah"
                attribute.  We will handle the element by doing an XPath search through
                the document, looking for the element with the matching id="blah"
                attribute.  We then recursively process that element after applying
                any necessary (x,y) translation.
                
                Notes:
                 1. We ignore the height and g attributes as they do not apply to
                    path-like elements, and
                 2. Even if the use element has visibility="hidden", SVG still calls
                    for processing the referenced element.  The referenced element is
                    hidden only if its visibility is "inherit" or "hidden".
                 3. We may be able to unlink clones using the code in pathmodifier.py
                """

                refid = node.get(inkex.addNS('href', 'xlink'))
                if refid is not None:
                    # [1:] to ignore leading '#' in reference
                    path = '//*[@id="{0}"]'.format(refid[1:])
                    refnode = node.xpath(path)
                    if refnode is not None:
                        x = float(node.get('x', '0'))
                        y = float(node.get('y', '0'))
                        # Note: the transform has already been applied
                        if x != 0 or y != 0:
                            mat_new2 = simpletransform.composeTransform(
                                matNew, simpletransform.parseTransform('translate({0:f},{1:f})'.format(x, y))
                            )
                        else:
                            mat_new2 = matNew
                        if len(refnode) > 0:
                            plottable, the_point = self.group_last_pt(refnode[0], mat_new2)
                        else:
                            plottable, the_point = self.group_last_pt(refnode, mat_new2)
                        return plottable, the_point
        except:
            pass

        # Svg Object is not a plottable element;
        # Return False and a default point
        return False, point

    def group_first_pt(self, group, matCurrent=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]):
        """
        Input: A Node which we have found to be a group
        Output: Boolean value to indicate if a point is plottable
                float values for first x,y coordinates of svg element
        """

        if len(group) == 0:  # Empty group -- The object may not be a group.
            return self.getFirstPoint(group, matCurrent)

        success = False
        point = [float(-1), float(-1)]

        # first apply the current matrix transform to this node's transform
        matNew = simpletransform.composeTransform(matCurrent, simpletransform.parseTransform(group.get('transform')))

        # Step through the group, we examine each element until we find a plottable object
        for subnode in group:
            # Check to see if the subnode we are looking at in this iteration of our for loop is a group
            # If it is a group, we must recursively call this function to search for a plottable object
            if subnode.tag == inkex.addNS('g', 'svg') or subnode.tag == 'g':
                # Verify that the nested group has objects within it
                # otherwise we will not parse it
                if subnode is not None:
                    # Check if group contains plottable elements by recursively calling group_first_pt
                    # If group contains plottable subnode, then it will return that value and escape the loop
                    # Else function continues search for first plottable object
                    success, point = self.group_first_pt(subnode, matNew)
                    if success:
                        # Subnode inside nested group is plottable!
                        # Break from our loop so we can return the first point of this plottable subnode
                        break
                    else:
                        continue
            else:
                # Node is not a group
                # Get its first (x,y) coordinates
                # Also get a Boolean value to indicate if the subnode is plottable or not
                # If subnode is not plottable, continue to next subnode in the group
                success, point = self.getFirstPoint(subnode, matNew)

                if success:
                    # Subnode inside group is plottable!
                    # Break from our loop so we can return the first point of this plottable subnode
                    break
                else:
                    continue
        return success, point

    def group_last_pt(self, group, matCurrent=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]):
        """
        Input: A Node which we have found to be a group
        Output: The last node within the group which can be plotted
        """

        if len(group) == 0:  # Empty group -- Did someone send an object that isn't a group?
            return self.getLastPoint(group, matCurrent)

        success = False
        point = [float(-1), float(-1)]

        # first apply the current matrix transform to this node's transform
        matNew = simpletransform.composeTransform(matCurrent, simpletransform.parseTransform(group.get('transform')))

        # Step through the group, we examine each element until we find a plottable object
        for subnode in reversed(group):
            # Check to see if the subnode we are looking at in this iteration of our for loop is a group
            # If it is a group, we must recursively call this function to search for a plottable object
            if subnode.tag == inkex.addNS('g', 'svg') or subnode.tag == 'g':
                # Verify that the nested group has objects within it
                # otherwise we will not parse it
                if subnode is not None:
                    # Check if group contains plottable elements by recursively calling group_last_pt
                    # If group contains plottable subnode, then it will return that value and escape the loop
                    # Else function continues search for last plottable object
                    success, point = self.group_last_pt(subnode, matNew)
                    if success:
                        # Subnode inside nested group is plottable!
                        # Break from our loop so we can return the first point of this plottable subnode
                        break
                    else:
                        continue
            else:
                # Node is not a group
                # Get its first (x,y) coordinates
                # Also get a Boolean value to indicate if the subnode is plottable or not
                # If subnode is not plottable, continue to next subnode in the group
                success, point = self.getLastPoint(subnode, matNew)
                if success:
                    # Subode inside nested group is plottable!
                    # Break from our loop so we can return the first point of this plottable subnode
                    break
                else:
                    continue
        return success, point

    def group2NodeDict(self, group, mat_current=None):
        if mat_current is None:
            mat_current = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]

        # first apply the current matrix transform to this node's transform
        matNew = simpletransform.composeTransform(mat_current, simpletransform.parseTransform(group.get('transform')))

        nodes_in_group = []

        # Step through the group, we examine each element until we find a plottable object
        for subnode in group:
            # Check to see if the subnode we are looking at in this iteration of our for loop is a group
            # If it is a group, we must recursively call this function to search for a plottable object
            if subnode.tag == inkex.addNS('g', 'svg') or subnode.tag == 'g':
                # Verify that the nested group has objects within it
                # otherwise we will not parse it
                if subnode is not None:
                    # Check if group contains plottable elements by recursively calling group_first_pt
                    # If group contains plottable subnode, then it will return that value and escape the loop
                    # Else function continues search for first plottable object
                    nodes_in_group.extend(self.group2NodeDict(subnode, matNew))
            else:
                simpletransform.applyTransformToNode(matNew, subnode)
                nodes_in_group.append(subnode)
        return nodes_in_group

    def ReorderNodeList(self, coord_dict, group_dict):
        # Re-order the given set of SVG elements, using a simple "greedy" algorithm.
        # The first object will be the element closest to the origin
        # After this choice, the algorithm loops through all remaining elements looking for the element whose first x,y
        # coordinates are closest to the the previous choice's last x,y coordinates
        # This process continues until all elements have been sorted into ordered_element_list and removed from group_dict

        ordered_layer_element_list = []

        # Continue until all elements have been re-ordered
        while group_dict:
            nearest_dist = float('inf')
            for key, node in group_dict.items():
                # Is this node non-plottable?
                # If so, exit loop and append element to ordered_layer_element_list
                if not coord_dict[key][0]:
                    # Object is not Plottable
                    nearest = node
                    nearest_id = key
                    continue

                # If we reach this point, node is plottable and needs to be considered in our algo
                entry_x = coord_dict[key][1]  # x-coordinate of first point of the path
                entry_y = coord_dict[key][2]  # y-coordinate of first point of the path

                exit_x = coord_dict[key][3]  # x-coordinate of last point of the path
                exit_y = coord_dict[key][4]  # y-coordinate of last point of the path

                object_dist = (entry_x - self.x_last) * (entry_x - self.x_last) + (entry_y - self.y_last) * (
                    entry_y - self.y_last
                )
                # This is actually the distance squared; calculating it rather than the pythagorean distance
                #  saves a square root calculation. Right now, we only care about _which distance is less_
                #  not the exact value of it, so this is a harmless shortcut.
                # If this distance is smaller than the previous element's distance, then replace the previous
                # element's entry with our current element's distance
                if nearest_dist >= object_dist:
                    # We have found an element closer than the previous closest element
                    nearest = node
                    nearest_id = key
                    nearest_dist = object_dist
                    nearest_start_x = entry_x
                    nearest_start_y = entry_y

            # Now that the closest object has been determined, it is time to add it to the
            # optimized list of closest objects
            ordered_layer_element_list.append(nearest)

            # To determine the closest object in the next iteration of the loop,
            # we must save the last x,y coor of this element
            # If this element is plottable, then save the x,y coordinates
            # If this element is non-plottable, then do not save the x,y coordinates
            if coord_dict[nearest_id][0]:
                # Also, draw line indicating that we've found a new point.
                if self.preview_rendering:
                    preview_path = []  # pen-up path data for preview

                    preview_path.append('M{0:.3f} {1:.3f}'.format(self.x_last, self.y_last))
                    preview_path.append('{0:.3f} {1:.3f}'.format(nearest_start_x, nearest_start_y))
                    self.p_style.update({'stroke': self.color_index(self.layer_index)})
                    path_attrs = {'style': simplestyle.formatStyle(self.p_style), 'd': ' '.join(preview_path)}

                    etree.SubElement(self.preview_layer, inkex.addNS('path', 'svg '), path_attrs, nsmap=inkex.NSS)

                self.x_last = coord_dict[nearest_id][3]
                self.y_last = coord_dict[nearest_id][4]

            # Remove this element from group_dict to indicate it has been optimized
            del group_dict[nearest_id]

        # Once all elements have been removed from the group_dictionary
        # Return the optimized list of svg elements in the layer
        return ordered_layer_element_list

    def color_index(self, index):
        index = index % 9

        if index == 0:
            return 'rgb(255, 0, 0))'
        elif index == 1:
            return 'rgb(170, 85, 0))'
        elif index == 2:
            return 'rgb(85, 170, 0))'
        elif index == 3:
            return 'rgb(0, 255, 0))'
        elif index == 4:
            return 'rgb(0, 170, 85))'
        elif index == 5:
            return 'rgb(0, 85, 170))'
        elif index == 6:
            return 'rgb(0, 0, 255))'
        elif index == 7:
            return 'rgb(85, 0, 170))'
        else:
            return 'rgb(170, 0, 85))'

    def getDocProps(self):
        """
        Get the document's height and width attributes from the <svg> tag.
        Use a default value in case the property is not present or is
        expressed in units of percentages.
        """

        self.svg_height = plot_utils.getLengthInches(self, 'height')
        self.svg_width = plot_utils.getLengthInches(self, 'width')

        width_string = self.svg.get('width')
        if width_string:
            value, units = plot_utils.parseLengthWithUnits(width_string)
            self.doc_units = units

        if self.auto_rotate and (self.svg_height > self.svg_width):
            self.printPortrait = True
        if self.svg_height is None or self.svg_width is None:
            return False
        else:
            return True

    def get_output(self):
        # Return serialized copy of svg document output
        result = etree.tostring(self.document)
        return result.decode('utf-8')


# Create effect instance and apply it.

if __name__ == '__main__':
    effect = ReorderEffect()
    exit_status.run(effect.affect)



# Copyright 2023 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
axidraw.py

Part of the AxiDraw driver for Inkscape
https://github.com/evil-mad/AxiDraw

See version_string below for current version and date.

Requires Python 3.8 or newer
"""
# pylint: disable=pointless-string-statement

__version__ = '3.9.6'  # Dated 2023-12-12

import copy
import gettext
import logging
import math
import socket  # for exception handling only
import time

import requests
import urllib3  # for exception handling only
from ink_extensions import inkex, simpletransform
from ink_extensions_utils import exit_status, message
from lxml import etree
from plotink import (
    ebb_motion,
    ebb_serial,  # https://github.com/evil-mad/plotink
    plot_utils,
    text_utils,
)

from . import (
    boundsclip,
    digest_svg,
    dripfeed,
    motion,
    path_objects,
    pen_handling,
    plot_optimizations,
    plot_status,
    plot_warnings,
    preview,
    serial_utils,
)
from .axidraw_options import common_options, versions

logger = logging.getLogger(__name__)


class AxiDraw(inkex.Effect):
    """Main class for AxiDraw"""

    logging_attrs = {'default_handler': message.UserMessageHandler()}

    def __init__(self, default_logging=True, user_message_fun=message.emit, params=None):
        if params is None:
            from . import axidraw_conf as params
            # params = import_module('axidrawinternal.axidraw_conf')  # Default configuration file
        self.params = params

        # axidraw.py is never actually called as a commandline tool, so why add options to
        # self.arg_parser here? Because it helps populate the self.options object
        # (argparse.Namespace) with necessary attributes and set the right defaults.
        # See self.initialize_options
        core_axidraw_options = common_options.core_axidraw_options(params.__dict__)
        inkex.Effect.__init__(self, common_options=[core_axidraw_options])

        self.initialize_options()

        self.version_string = __version__

        self.plot_status = plot_status.PlotStatus()
        self.pen = pen_handling.PenHandler()
        self.warnings = plot_warnings.PlotWarnings()
        self.preview = preview.Preview()

        self.spew_debugdata = False  # Possibly add this as a PlotStatus variable
        self.set_defaults()
        self.digest = None
        self.vb_stash = [1, 1, 0, 0]  # Viewbox storage
        self.bounds = [[0, 0], [0, 0]]
        self.connected = False  # Python API variable.

        self.plot_status.secondary = False
        self.user_message_fun = user_message_fun

        if default_logging:  # logging setup
            logger.setLevel(logging.INFO)
            logger.addHandler(self.logging_attrs['default_handler'])

        if self.spew_debugdata:
            logger.setLevel(logging.DEBUG)  # by default level is INFO

    def set_up_pause_receiver(self, software_pause_event):
        """use a multiprocessing.Event/threading.Event to communicate a
        keyboard interrupt (ctrl-C) to pause the AxiDraw"""
        self._software_pause_event = software_pause_event

    def receive_pause_request(self):
        """pause receiver"""
        return hasattr(self, '_software_pause_event') and self._software_pause_event.is_set()

    def set_secondary(self, suppress_standard_out=True):
        """If a "secondary" AxiDraw called by axidraw_control"""
        self.plot_status.secondary = True
        self.called_externally = True
        if suppress_standard_out:
            self.suppress_standard_output_stream()

    def suppress_standard_output_stream(self):
        """Save values we will need later in unsuppress_standard_output_stream"""
        self.logging_attrs['additional_handlers'] = [SecondaryErrorHandler(self), SecondaryNonErrorHandler(self)]
        self.logging_attrs['emit_fun'] = self.user_message_fun
        logger.removeHandler(self.logging_attrs['default_handler'])
        for handler in self.logging_attrs['additional_handlers']:
            logger.addHandler(handler)

    def unsuppress_standard_output_stream(self):
        """Release logging stream"""
        logger.addHandler(self.logging_attrs['default_handler'])
        if self.logging_attrs['additional_handlers']:
            for handler in self.logging_attrs['additional_handlers']:
                logger.removeHandler(handler)

        self.user_message_fun = self.logging_attrs['emit_fun']

    def set_defaults(self):
        """Set default values of certain parameters
        These are set when the class is initialized.
        Also called in plot_run() in the Python API, to ensure that
        these defaults are set before plotting additional pages."""

        self.use_layer_speed = False
        self.plot_status.reset()  # Clear serial port and pause status flags
        self.pen.reset()  # Clear pen state, lift count, layer pen height flag
        self.warnings.reset()  # Clear any warning messages
        self.time_elapsed = 0  # Available for use by python API

        self.svg_transform = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]
        self.digest = None

    def initialize_options(self):
        """Use the flags and arguments defined in __init__ to populate self.options with
        the necessary attributes and set defaults"""
        self.getoptions([])
        # self.getoptions initializes self.options by calling self.arg_parser.parse_args, which is
        # not the intended use of parse_args

    def update_options(self):
        """Parse and update certain options; called in effect and in interactive modes
        whenever the options are updated"""

        x_bounds_min = 0.0
        y_bounds_min = 0.0

        # Physical travel bounds, based on AxiDraw model:
        if self.options.model == 2:
            x_bounds_max = self.params.x_travel_V3A3
            y_bounds_max = self.params.y_travel_V3A3
        elif self.options.model == 3:
            x_bounds_max = self.params.x_travel_V3XLX
            y_bounds_max = self.params.y_travel_V3XLX
        elif self.options.model == 4:
            x_bounds_max = self.params.x_travel_MiniKit
            y_bounds_max = self.params.y_travel_MiniKit
        elif self.options.model == 5:
            x_bounds_max = self.params.x_travel_SEA1
            y_bounds_max = self.params.y_travel_SEA1
        elif self.options.model == 6:
            x_bounds_max = self.params.x_travel_SEA2
            y_bounds_max = self.params.y_travel_SEA2
        elif self.options.model == 7:
            x_bounds_max = self.params.x_travel_V3B6
            y_bounds_max = self.params.y_travel_V3B6
        else:
            x_bounds_max = self.params.x_travel_default
            y_bounds_max = self.params.y_travel_default

        self.bounds = [[x_bounds_min - 1e-9, y_bounds_min - 1e-9], [x_bounds_max + 1e-9, y_bounds_max + 1e-9]]

        # Speeds in inches/second:
        self.speed_pendown = self.params.speed_pendown * self.params.speed_lim_xy_hr / 110.0
        self.speed_penup = self.params.speed_penup * self.params.speed_lim_xy_hr / 110.0

        # Input limit checking; constrain input values and prevent zero speeds:
        self.options.pen_pos_up = plot_utils.constrainLimits(self.options.pen_pos_up, 0, 100)
        self.options.pen_pos_down = plot_utils.constrainLimits(self.options.pen_pos_down, 0, 100)
        self.options.pen_rate_raise = plot_utils.constrainLimits(self.options.pen_rate_raise, 1, 200)
        self.options.pen_rate_lower = plot_utils.constrainLimits(self.options.pen_rate_lower, 1, 200)
        self.options.speed_pendown = plot_utils.constrainLimits(self.options.speed_pendown, 1, 110)
        self.options.speed_penup = plot_utils.constrainLimits(self.options.speed_penup, 1, 200)
        self.options.accel = plot_utils.constrainLimits(self.options.accel, 1, 110)

    def effect(self):
        """Main entry point: check to see which mode/tab is selected, and act accordingly."""
        self.start_time = time.time()

        try:
            self.plot_status.secondary
        except AttributeError:
            self.plot_status.secondary = False

        self.text_out = ''  # Text log for basic communication messages
        self.error_out = ''  # Text log for significant errors

        self.plot_status.stats.reset()  # Reset plot duration and distance statistics

        self.doc_units = 'in'

        self.pen.phys.xpos = self.params.start_pos_x
        self.pen.phys.ypos = self.params.start_pos_y

        self.layer_speed_pendown = -1
        self.plot_status.copies_to_plot = 1

        self.plot_status.resume.reset()  # New values to write to file:

        self.svg_width = 0
        self.svg_height = 0
        self.rotate_page = False

        self.update_options()

        self.options.mode = self.options.mode.strip('"')  # Input sanitization
        self.options.setup_type = self.options.setup_type.strip('"')
        self.options.manual_cmd = self.options.manual_cmd.strip('"')
        self.options.resume_type = self.options.resume_type.strip('"')
        self.options.page_delay = max(self.options.page_delay, 0)

        try:
            self.called_externally
        except AttributeError:
            self.called_externally = False

        if self.options.mode == 'options':
            return
        if self.options.mode == 'timing':
            return
        if self.options.mode == 'version':
            # Return the version of _this python script_.
            self.user_message_fun(self.version_string)
            return
        if self.options.mode == 'manual':
            if self.options.manual_cmd == 'none':
                return  # No option selected. Do nothing and return no error.
            if self.options.manual_cmd == 'strip_data':
                self.svg = self.document.getroot()
                for slug in ['WCB', 'MergeData', 'plotdata', 'eggbot']:
                    for node in self.svg.xpath('//svg:' + slug, namespaces=inkex.NSS):
                        self.svg.remove(node)
                self.user_message_fun(gettext.gettext('All AxiDraw data has been removed from this SVG file.'))
                return
            if self.options.manual_cmd in ('res_read', 'res_adj_in', 'res_adj_mm'):
                self.svg = self.document.getroot()
                self.user_message_fun(self.plot_status.resume.manage_offset(self))
                self.res_dist = max(self.plot_status.resume.new.pause_dist * 25.4, 0)  # Python API
                return
            if self.options.manual_cmd == 'list_names':  # Run before regular serial connection!
                self.name_list = ebb_serial.list_named_ebbs()  # Variable available for python API
                if not self.name_list:
                    self.user_message_fun(gettext.gettext('No named AxiDraw units located.\n'))
                else:
                    self.user_message_fun(gettext.gettext('List of attached AxiDraw units:'))
                    for detected_ebb in self.name_list:
                        self.user_message_fun(detected_ebb)
                return

        if self.options.mode == 'resume':
            if self.options.resume_type == 'home':
                self.options.mode = 'res_home'
            else:
                self.options.mode = 'res_plot'
                self.options.copies = 1

        if self.options.mode == 'setup':
            # setup mode -> either align, toggle, or cycle modes.
            self.options.mode = self.options.setup_type

        if self.options.digest > 1:  # Generate digest only; do not run plot or preview
            self.options.preview = True  # Disable serial communication; restrict certain functions

        if not self.options.preview:
            self.serial_connect()
            self.plot_status.resume.clear_button(self)  # Query button to clear its state

        if self.options.mode == 'sysinfo':
            versions.report_version_info(
                self.plot_status,
                self.params.check_updates,
                self.version_string,
                self.options.preview,
                self.user_message_fun,
            )

        if self.plot_status.port is None and not self.options.preview:
            return  # unable to connect to axidraw

        if self.options.mode in ('align', 'toggle', 'cycle'):
            self.setup_command()
            self.warnings.report(self.called_externally, self.user_message_fun)  # print warnings
            return

        if self.options.mode == 'manual':
            self.manual_command()  # Handle manual commands that use both power and usb.
            self.warnings.report(self.called_externally, self.user_message_fun)  # print warnings
            return

        self.svg = self.document.getroot()
        self.plot_status.resume.update_needed = False
        self.plot_status.resume.new.model = self.options.model  # Save model in file

        if self.options.mode in ('plot', 'layers', 'res_plot', 'res_home'):
            # Read saved data from SVG file, including plob version information
            self.plot_status.resume.read_from_svg(self.svg)

        if self.options.mode == 'res_plot':  # Initialization for resuming plots
            if self.plot_status.resume.old.pause_dist >= 0:
                self.pen.phys.xpos = self.plot_status.resume.old.last_x
                self.pen.phys.ypos = self.plot_status.resume.old.last_y
                self.plot_status.resume.new.rand_seed = self.plot_status.resume.old.rand_seed
                self.plot_status.resume.new.layer = self.plot_status.resume.old.layer
            else:
                logger.error(gettext.gettext('No in-progress plot data found in file; unable to resume.'))
                return

        if self.options.mode in ('plot', 'layers', 'res_plot'):
            self.plot_status.copies_to_plot = self.options.copies
            if self.plot_status.copies_to_plot == 0:  # Special case: Continuous copies selected
                self.plot_status.copies_to_plot = -1  # Flag for continuous copies

            if self.options.preview and not self.options.random_start:
                # Special preview case: Without randomizing, pages have identical print time:
                self.plot_status.copies_to_plot = 1

            if self.options.mode == 'plot':
                self.plot_status.resume.new.layer = -1  # Plot all layers
            if self.options.mode == 'layers':
                self.plot_status.resume.new.layer = self.options.layer

            # Parse & digest SVG document, perform initial optimizations, prepare to resume:
            if not self.prepare_document():
                return

            if self.options.digest > 1:  # Generate digest only; do not run plot or preview
                self.plot_cleanup()  # Revert document to save plob & print time elapsed
                self.plot_status.resume.new.plob_version = str(path_objects.PLOB_VERSION)
                self.plot_status.resume.write_to_svg(self.svg)
                self.warnings.report(False, self.user_message_fun)  # print warnings
                return

            if self.options.mode == 'res_plot':  # Crop digest up to when the plot resumes:
                self.digest.crop(self.plot_status.resume.old.pause_dist)

            # CLI PROGRESS BAR: SET UP DRY RUN TO ESTIMATE PLOT LENGTH & TIME
            if self.plot_status.progress.review(self.plot_status, self.options):
                self.plot_document()  # "Dry run": Estimate plot length & time

                self.user_message_fun(self.plot_status.progress.restore(self))
                self.plot_status.stats.reset()  # Reset plot duration and distance statistics

            if self.options.mode == 'res_plot':
                self.pen.phys.xpos = self.plot_status.resume.old.last_x
                self.pen.phys.ypos = self.plot_status.resume.old.last_y

                # Update so that if the plot is paused, we can resume again
                self.plot_status.stats.down_travel_inch = self.plot_status.resume.old.pause_dist

            first_copy = True
            while self.plot_status.copies_to_plot != 0:
                self.preview.reset()  # Clear preview data before starting each plot
                self.plot_status.resume.update_needed = True
                self.plot_status.copies_to_plot -= 1

                if first_copy:
                    first_copy = False
                else:
                    self.plot_status.stats.next_page()  # Update distance stats for next page
                    if self.options.random_start:
                        self.randomize_optimize()  # Only need to re-optimize if randomizing
                self.plot_document()
                dripfeed.page_layer_delay(self, between_pages=True)  # Delay between pages

            self.plot_cleanup()  # Revert document, print time reports, send webhooks

        elif self.options.mode == 'res_home':
            self.plot_status.resume.copy_old()
            self.pen.phys.xpos = self.plot_status.resume.old.last_x
            self.pen.phys.ypos = self.plot_status.resume.old.last_y
            self.plot_status.resume.update_needed = True

            if not self.plot_status.resume.read:
                logger.error(gettext.gettext('No resume data found; unable to return Home.'))
                return
            if math.fabs(self.pen.phys.xpos < 0.001) and math.fabs(self.pen.phys.ypos < 0.001):
                logger.error(gettext.gettext('Unable to move to Home. (Is the AxiDraw already at Home?)'))
                return

            self.query_ebb_voltage()
            self.pen.servo_init(self)
            self.pen.pen_raise(self)
            self.enable_motors()
            self.go_to_position(self.params.start_pos_x, self.params.start_pos_y)

        if self.plot_status.resume.update_needed:
            self.plot_status.resume.new.last_x = self.pen.phys.xpos
            self.plot_status.resume.new.last_y = self.pen.phys.ypos
            if self.options.digest:  # i.e., if self.options.digest > 0
                self.plot_status.resume.new.plob_version = str(path_objects.PLOB_VERSION)
            self.plot_status.resume.write_to_svg(self.svg)
        if self.plot_status.port is not None:
            ebb_motion.doTimedPause(self.plot_status.port, 10, False)  # Final timed motion command
            if self.options.port is None:  # Do not close serial port if it was opened externally.
                self.disconnect()
        self.warnings.report(self.called_externally, self.user_message_fun)  # print warnings

    def setup_command(self):
        """Commands from the setup modes. Need power and USB, but not SVG file."""

        if self.options.preview:
            self.user_message_fun('Command unavailable while in preview mode.')
            return

        if self.plot_status.port is None:
            return

        self.query_ebb_voltage()
        self.pen.servo_init(self)

        if self.options.mode == 'align':
            self.pen.pen_raise(self)
            ebb_motion.sendDisableMotors(self.plot_status.port, False)
        elif self.options.mode == 'cycle':
            self.pen.cycle(self)
        # Note that "toggle" mode is handled within self.pen.servo_init(self)

    def manual_command(self):
        """Manual mode commands that need USB connectivity and don't need SVG file"""

        if self.options.preview:  # First: Commands that require serial but not power
            self.user_message_fun('Command unavailable while in preview mode.')
            return
        if self.plot_status.port is None:
            return

        if self.options.manual_cmd == 'fw_version':
            self.user_message_fun(self.plot_status.fw_version)
            return

        if self.options.manual_cmd == 'bootload':
            success = ebb_serial.bootload(self.plot_status.port)
            if success:
                self.user_message_fun(
                    gettext.gettext(
                        'Entering bootloader mode for firmware programming.\n'
                        + 'To resume normal operation, you will need to first\n'
                        + 'disconnect the AxiDraw from both USB and power.'
                    )
                )
                self.disconnect()  # Disconnect from AxiDraw; end serial session
            else:
                logger.error('Failed while trying to enter bootloader.')
            return

        if self.options.manual_cmd == 'read_name':
            name_string = ebb_serial.query_nickname(self.plot_status.port)
            if name_string is None:
                logger.error(gettext.gettext('Error; unable to read nickname.\n'))
            else:
                self.user_message_fun(name_string)
            return

        if (self.options.manual_cmd).startswith('write_name'):
            temp_string = self.options.manual_cmd
            temp_string = temp_string.split('write_name', 1)[1]  # Get part after "write_name"
            temp_string = temp_string[:16]  # Only use first 16 characters in name
            if not temp_string:
                temp_string = ''  # Use empty string to clear nickname.

            if versions.min_fw_version(self.plot_status, '2.5.5'):
                renamed = ebb_serial.write_nickname(self.plot_status.port, temp_string)
                if renamed is True:
                    if temp_string == '':
                        self.user_message_fun('Writing "blank" Nickname; setting to default.')
                    else:
                        self.user_message_fun(f'Nickname "{temp_string}" written.')
                    self.user_message_fun('Rebooting EBB.')
                else:
                    logger.error('Error encountered while writing nickname.')
                ebb_serial.reboot(self.plot_status.port)  # Reboot required after writing nickname
                self.disconnect()  # Disconnect from AxiDraw; end serial session
            else:
                logger.error('This function requires a newer firmware version. See: axidraw.com/fw')
            return

        self.query_ebb_voltage()  # Next: Commands that also require both power to move motors:
        if self.options.manual_cmd == 'raise_pen':
            self.pen.servo_init(self)  # Initializes to pen-up position
        elif self.options.manual_cmd == 'lower_pen':
            self.pen.servo_init(self)  # Initializes to pen-down position
        elif self.options.manual_cmd == 'enable_xy':
            self.enable_motors()
        elif self.options.manual_cmd == 'disable_xy':
            ebb_motion.sendDisableMotors(self.plot_status.port, False)
        else:  # walk motors or move home cases:
            self.pen.servo_init(self)
            self.enable_motors()  # Set plotting resolution
            if self.options.manual_cmd == 'walk_home':
                if versions.min_fw_version(self.plot_status, '2.6.2'):
                    serial_utils.exhaust_queue(self)  # Wait until all motion stops
                    a_pos, b_pos = ebb_motion.query_steps(self.plot_status.port, False)
                    n_delta_x = -(a_pos + b_pos) / (4 * self.params.native_res_factor)
                    n_delta_y = -(a_pos - b_pos) / (4 * self.params.native_res_factor)
                    if self.options.resolution == 2:  # Low-resolution mode
                        n_delta_x *= 2
                        n_delta_y *= 2
                    ### << CRITICAL >> aligns interactive and plot modes:
                    self.pen.phys.xpos = -n_delta_x
                    self.pen.phys.ypos = -n_delta_y
                else:
                    logger.error('This function requires newer firmware. Update at: axidraw.com/fw')
                    return
            elif self.options.manual_cmd == 'walk_y':
                n_delta_x = 0
                n_delta_y = self.options.dist
            elif self.options.manual_cmd == 'walk_x':
                n_delta_y = 0
                n_delta_x = self.options.dist
            elif self.options.manual_cmd == 'walk_mmy':
                n_delta_x = 0
                n_delta_y = self.options.dist / 25.4
            elif self.options.manual_cmd == 'walk_mmx':
                n_delta_y = 0
                n_delta_x = self.options.dist / 25.4
            else:
                return
            f_x = self.pen.phys.xpos + n_delta_x  # Note: Walks are relative, not absolute!
            f_y = self.pen.phys.ypos + n_delta_y  # New position is not saved; use with care.
            self.go_to_position(f_x, f_y, ignore_limits=True)

    def prepare_document(self):
        """
        Prepare the SVG document for plotting: Create the plot digest, join nearby ends,
        and perform supersampling. If not using randomization, then optimize the digest as well.
        """
        if not self.get_doc_props():
            logger.error(gettext.gettext('This document does not have valid dimensions.'))
            logger.error(gettext.gettext('The page size should be in either millimeters (mm) or inches (in).\r\r'))
            logger.error(gettext.gettext('Consider starting with the Letter landscape or '))
            logger.error(gettext.gettext('the A4 landscape template.\r\r'))
            logger.error(gettext.gettext('The page size may also be set in Inkscape,\r'))
            logger.error(gettext.gettext('using File > Document Properties.'))
            return False

        if not hasattr(self, 'backup_original'):
            self.backup_original = copy.deepcopy(self.document)

        # Modifications to SVG -- including re-ordering and text substitution
        #   may be made at this point, and will not be preserved.

        v_b = self.svg.get('viewBox')
        if v_b:
            p_a_r = self.svg.get('preserveAspectRatio')
            s_x, s_y, o_x, o_y = plot_utils.vb_scale(v_b, p_a_r, self.svg_width, self.svg_height)
        else:
            s_x = 1.0 / float(plot_utils.PX_PER_INCH)  # Handle case of no viewbox
            s_y = s_x
            o_x = 0.0
            o_y = 0.0
        self.vb_stash = s_x, s_y, o_x, o_y

        # Initial transform of document is based on viewbox, if present:
        self.svg_transform = simpletransform.parseTransform(
            f'scale({s_x:.6E},{s_y:.6E}) translate({o_x:.6E},{o_y:.6E})'
        )

        valid_plob = False
        if self.plot_status.resume.old.plob_version:
            logger.debug('Checking Plob')
            valid_plob = digest_svg.verify_plob(self.svg, self.options.model)
        if valid_plob:
            logger.debug('Valid plob found; skipping standard pre-processing.')
            self.digest = path_objects.DocDigest()
            self.digest.from_plob(self.svg)
            self.plot_status.resume.new.plob_version = str(path_objects.PLOB_VERSION)
        else:  # Process the input SVG into a simplified, restricted-format DocDigest object:
            digester = digest_svg.DigestSVG()  # Initialize class
            if self.options.hiding:  # Process all visible layers
                digest_params = [self.svg_width, self.svg_height, s_x, s_y, -2, self.params.curve_tolerance]
            else:  # Process only selected layer, if in layers mode
                digest_params = [
                    self.svg_width,
                    self.svg_height,
                    s_x,
                    s_y,
                    self.plot_status.resume.new.layer,
                    self.params.curve_tolerance,
                ]
            self.digest = digester.process_svg(
                self.svg,
                self.warnings,
                digest_params,
                self.svg_transform,
            )

            if self.rotate_page:  # Rotate digest
                self.digest.rotate(self.params.auto_rotate_ccw)

            if self.options.hiding:
                """
                Perform hidden-line clipping at this point, based on object
                    fills, clipping masks, and document and plotting bounds, via self.bounds
                """
                # clipping involves a non-pure Python dependency (pyclipper), so only import
                # when necessary
                from axidrawinternal.clipping import ClipPathsProcess

                bounds = ClipPathsProcess.calculate_bounds(
                    self.bounds, self.svg_height, self.svg_width, self.params.clip_to_page, self.rotate_page
                )
                # flattening removes essential information for the clipping process
                assert not self.digest.flat
                self.digest.layers = ClipPathsProcess().run(self.digest.layers, bounds, clip_on=True)
                self.digest.layer_filter(self.plot_status.resume.new.layer)  # For Layers mode
                self.digest.remove_unstroked()  # Only stroked objects can plot
                self.digest.flatten()  # Flatten digest before optimizations and plotting
            else:
                """
                Clip digest at plot bounds
                """
                if self.rotate_page:
                    doc_bounds = [self.svg_height + 1e-9, self.svg_width + 1e-9]
                else:
                    doc_bounds = [self.svg_width + 1e-9, self.svg_height + 1e-9]
                out_of_bounds_flag = boundsclip.clip_at_bounds(
                    self.digest, self.bounds, doc_bounds, self.params.bounds_tolerance, self.params.clip_to_page
                )
                if out_of_bounds_flag:
                    self.warnings.add_new('bounds')

            """
            Possible future work: Perform automatic hatch filling at this point, based on object
                fill colors and possibly other factors.
            """

            """
            Optimize digest
            """

            allow_reverse = self.options.reordering in [2, 3]

            if self.options.reordering < 3:  # Set reordering to 4 to disable path joining
                plot_optimizations.connect_nearby_ends(self.digest, allow_reverse, self.params.min_gap)

            plot_optimizations.supersample(self.digest, self.params.segment_supersample_tolerance)

            self.randomize_optimize(True)  # Do plot randomization & optimizations

        # If it is necessary to save as a Plob, that conversion can be made like so:
        # plob = self.digest.to_plob() # Unnecessary re-conversion for testing only
        # self.digest.from_plob(plob)  # Unnecessary re-conversion for testing only
        return True

    def randomize_optimize(self, first_copy=False):
        """Randomize start points & perform reordering"""

        if self.plot_status.resume.new.plob_version != 'n/a':
            return  # Working from valid plob; do not perform any optimizations.
        if self.options.random_start:
            if self.options.mode != 'res_plot':  # Use old rand seed when resuming a plot.
                self.plot_status.resume.new.rand_seed = int(time.time() * 100)
            plot_optimizations.randomize_start(self.digest, self.plot_status.resume.new.rand_seed)

        allow_reverse = self.options.reordering in [2, 3]

        if self.options.reordering in [1, 2, 3]:
            plot_optimizations.reorder(self.digest, allow_reverse)

        if first_copy and self.options.digest:  # Will return Plob, not full SVG; back it up here.
            self.backup_original = copy.deepcopy(self.digest.to_plob())

    def plot_document(self):
        """Plot the prepared SVG document, if so selected in the interface"""

        if not self.options.preview:
            self.plot_status.resume.clear_button(self)  # Query button to clear its state
            self.options.rendering = 0  # Only render previews if we are in preview mode.
            self.preview.v_chart.enable = False
            if self.plot_status.port is None:
                return
            self.query_ebb_voltage()

        self.plot_status.progress.launch(self)

        try:  # wrap everything in a try so we can be sure to close the serial port
            self.pen.servo_init(self)
            self.pen.pen_raise(self)
            self.enable_motors()  # Set plotting resolution

            self.plot_doc_digest(self.digest)  # Step through and plot contents of document digest
            self.pen.pen_raise(self)

            if self.plot_status.stopped == 0:  # Return Home after normal plot
                self.plot_status.resume.new.clean()  # Clear flags indicating resume status
                self.go_to_position(self.params.start_pos_x, self.params.start_pos_y)

        finally:  # In case of an exception and loss of the serial port...
            pass

        self.plot_status.progress.close()

    def plot_cleanup(self):
        """
        Perform standard actions after a plot or the last copy from a set of plots:
        Revert file, render previews, print time reports, run webhook.

        Reverting is back to original SVG document, prior to adding preview layers.
            and prior to saving updated "plotdata" progress data in the file.
            No changes to the SVG document prior to this point will be saved.

        Doing so allows us to use routines that alter the SVG prior to this point,
            e.g., plot re-ordering for speed or font substitutions.
        """
        self.document = copy.deepcopy(self.backup_original)

        try:  # Handle cases: backup_original May be etree Element or ElementTree
            self.svg = self.document.getroot()  # For ElementTree, get the root
        except AttributeError:
            self.svg = self.document  # For Element; no need to get the root

        if self.options.digest:
            self.options.rendering = 0  # Turn off rendering

        if self.options.digest > 1:  # Save Plob file only and exit.
            elapsed_time = time.time() - self.start_time
            self.time_elapsed = elapsed_time  # Available for use by python API
            if self.options.report_time and not self.called_externally:  # Print time only
                self.user_message_fun('Elapsed time: ' + text_utils.format_hms(elapsed_time))
            return

        self.preview.render(self)  # Render preview on the page, if enabled and in preview mode

        if self.plot_status.progress.enable and self.plot_status.stopped == 0:
            self.user_message_fun('\nAxiCLI plot complete.\n')  # If sequence ended normally.
        elapsed_time = time.time() - self.start_time
        self.time_elapsed = elapsed_time  # Available for use by python API

        if not self.called_externally:  # Compile time estimates & print time reports
            self.plot_status.stats.report(self.options, self.user_message_fun, elapsed_time)
            self.pen.status.report(self, self.user_message_fun)
            if self.options.report_time and self.plot_status.resume.new.plob_version != 'n/a':
                self.user_message_fun('Document printed from valid Plob digest.')

        if self.options.webhook and not self.options.preview:
            if self.options.webhook_url is not None:
                payload = {
                    'value1': str(self.digest.name),
                    'value2': str(text_utils.format_hms(elapsed_time)),
                    'value3': str(self.options.port),
                }
                try:
                    requests.post(self.options.webhook_url, data=payload, timeout=7)
                except (
                    TimeoutError,
                    urllib3.exceptions.ConnectTimeoutError,
                    urllib3.exceptions.MaxRetryError,
                    requests.exceptions.ConnectTimeout,
                ):
                    self.user_message_fun('Webhook notification failed (Timed out).\n')
                except (urllib3.exceptions.NewConnectionError, socket.gaierror, requests.exceptions.ConnectionError):
                    self.user_message_fun(
                        'An error occurred while posting webhook. '
                        + 'Check your internet connection and webhook URL.\n'
                    )

    def plot_doc_digest(self, digest):
        """
        Step through the document digest and plot each of the vertex lists.

        Takes a flattened path_objects.DocDigest object as input. All
        selection of elements to plot and their rendering, including
        transforms, needs to be handled before this routine.
        """

        if not digest:
            return

        for layer in digest.layers:
            self.pen.end_temp_height(self)
            old_use_layer_speed = self.use_layer_speed  # A Boolean
            old_layer_speed_pendown = self.layer_speed_pendown  # Numeric value
            self.pen.pen_raise(self)  # Raise pen prior to computing layer properties

            if self.options.mode == 'layers':  # Special case: The plob contains all layers
                if layer.props.number != self.options.layer:  # and is plotted in layers mode.
                    continue  # Here, ensure that only certain layers should be printed.

            self.eval_layer_props(layer.props)

            for path_item in layer.paths:
                if self.plot_status.stopped:
                    return
                self.plot_polyline(path_item.subpaths[0])
            self.use_layer_speed = old_use_layer_speed  # Restore old layer status variables

            if self.layer_speed_pendown != old_layer_speed_pendown:
                self.layer_speed_pendown = old_layer_speed_pendown
                self.enable_motors()  # Set speed value variables for this layer.
            self.pen.end_temp_height(self)

    def eval_layer_props(self, layer_props):
        """
        Check for encoded pause, delay, speed, or height in the layer name, and act upon them.
        Syntax described at: https://wiki.evilmadscientist.com/AxiDraw_Layer_Control
        """

        if layer_props.pause:  # Insert programmatic pause
            if not self.plot_status.progress.dry_run:  # Skip during dry run only
                if self.plot_status.stopped == 0:  # If not already stopped
                    self.plot_status.stopped = -1  # Set flag for programmatic pause
                self.pause_check()  # Carry out the pause, or resume if required.

        old_speed = self.layer_speed_pendown

        self.use_layer_speed = False
        self.layer_speed_pendown = -1

        if layer_props.delay:
            dripfeed.page_layer_delay(self, between_pages=False, delay_ms=layer_props.delay)
        if layer_props.height is not None:  # New height will be used when we next lower the pen.
            self.pen.set_temp_height(self, layer_props.height)
        if layer_props.speed:
            self.use_layer_speed = True
            self.layer_speed_pendown = layer_props.speed

        if self.layer_speed_pendown != old_speed:
            self.enable_motors()  # Set speed value variables for this layer.

    def plot_polyline(self, vertex_list):
        """
        Plot a polyline object; a single pen-down XY movement.
        - No transformations, no curves, no neat clipping at document bounds;
            those are all performed _before_ we get to this point.
        - Truncate motion, brute-force, at travel bounds, without mercy or printed warnings.
        """

        if self.plot_status.stopped:
            logger.debug('Polyline: self.plot_status.stopped.')
            return
        if not vertex_list:
            logger.debug('No vertex list to plot. Returning.')
            return
        if len(vertex_list) < 2:
            logger.debug('No full segments in vertex list. Returning.')
            return

        self.pen.pen_raise(self)  # Raise, if necessary, prior to pen-up travel to first vertex

        for vertex in vertex_list:
            vertex[0], _t_x = plot_utils.checkLimitsTol(vertex[0], 0, self.bounds[1][0], 2e-9)
            vertex[1], _t_y = plot_utils.checkLimitsTol(vertex[1], 0, self.bounds[1][1], 2e-9)
            # if _t_x or _t_y:
            #     logger.debug('Travel truncated to bounds at plot_polyline.')

        # Pen up straight move, zero velocity at endpoints, to first vertex location
        self.go_to_position(vertex_list[0][0], vertex_list[0][1])

        # Plan and feed trajectory, including lowering and raising pen before and after:
        the_trajectory = motion.trajectory(self, vertex_list)
        dripfeed.feed(self, the_trajectory[0])

    def go_to_position(self, x_dest, y_dest, ignore_limits=False, xyz_pos=None):
        """
        Immediate XY move to destination, using normal motion planning. Replaces legacy
        function "plot_seg_with_v", assuming zero initial and final velocities.
        """
        target_data = (x_dest, y_dest, 0, 0, ignore_limits)
        the_trajectory = motion.compute_segment(self, target_data, xyz_pos)
        dripfeed.feed(self, the_trajectory[0])

    def pause_check(self):
        """Manage Pause functionality and stop plot if requested or at certain errors"""
        if self.plot_status.stopped > 0:
            return  # Plot is already stopped. No need to proceed.

        pause_button_pressed = self.plot_status.resume.check_button(self)

        if self.receive_pause_request():  # Keyboard interrupt detected!
            self.plot_status.stopped = -103  # Code 104: "Keyboard interrupt"
            if self.plot_status.delay_between_copies:  # However... it could have been...
                self.plot_status.stopped = -2  # Paused between copies (OK).

        if self.plot_status.stopped == -1:
            self.user_message_fun('Plot paused programmatically.\n')
        if self.plot_status.stopped == -103:
            self.user_message_fun('\nPlot paused by keyboard interrupt.\n')

        if pause_button_pressed == -1:
            self.user_message_fun(
                '\nError: USB connection to AxiDraw lost. '
                + f'[Position: {25.4 * self.plot_status.stats.down_travel_inch:.3f} mm]\n'
            )

            self.connected = False  # Python interactive API variable
            self.plot_status.stopped = -104  # Code 104: "Lost connectivity"

        if pause_button_pressed == 1:
            if self.plot_status.delay_between_copies:
                self.plot_status.stopped = -2  # Paused between copies.
            elif self.options.mode == 'interactive':
                logger.warning('Plot halted by button press during interactive session.')
                logger.warning('Manually home this AxiDraw before plotting next item.\n')
                self.plot_status.stopped = -102  # Code 102: "Paused by button press"
            else:
                self.user_message_fun('Plot paused by button press.\n')
                self.plot_status.stopped = -102  # Code 102: "Paused by button press"

        if self.plot_status.stopped == -2:
            self.user_message_fun('Plot sequence ended between copies.\n')

        if self.plot_status.stopped in (-1, -102, -103):
            self.user_message_fun(
                '(Paused after: ' + f'{25.4 * self.plot_status.stats.down_travel_inch:.3f} mm of pen-down travel.)'
            )

        if self.plot_status.stopped < 0:  # Stop plot
            self.pen.pen_raise(self)
            if (
                not self.plot_status.delay_between_copies
                and not self.plot_status.secondary
                and self.options.mode != 'interactive'
            ):
                # Only print if we're not in the delay between copies, nor a "second" unit.
                if self.plot_status.stopped != -104:  # Do not display after loss of USB.
                    self.user_message_fun('Use the resume feature to continue.\n')
            self.plot_status.stopped = -self.plot_status.stopped
            self.plot_status.copies_to_plot = 0

            if self.options.mode not in ('plot', 'layers', 'res_plot'):
                return  # Don't update pause_dist in res_home or repositioning modes

            self.plot_status.resume.new.pause_dist = self.plot_status.stats.down_travel_inch
            self.plot_status.resume.new.pause_ref = self.plot_status.stats.down_travel_inch

    def serial_connect(self):
        """Connect to AxiDraw over USB"""
        if serial_utils.connect(self.options, self.plot_status, self.user_message_fun, logger):
            self.connected = True  # Variable available in the Python interactive API.
        else:
            self.plot_status.stopped = 101  # Will become exit code 101; failed to connect

    def enable_motors(self):
        """
        Enable motors, set native motor resolution, and set speed scales.
        The "pen down" speed scale is adjusted by reducing speed when using 8X microstepping or
        disabling aceleration. These factors prevent unexpected dramatic changes in speed when
        turning those two options on and off.
        """
        if self.use_layer_speed:
            local_speed_pendown = self.layer_speed_pendown
        else:
            local_speed_pendown = self.options.speed_pendown

        if self.options.resolution == 1:  # High-resolution ("Super") mode
            if not self.options.preview:
                res_1, res_2 = ebb_motion.query_enable_motors(self.plot_status.port, False)
                if not (res_1 == 1 and res_2 == 1):  # Do not re-enable if already enabled
                    ebb_motion.sendEnableMotors(self.plot_status.port, 1)  # 16X microstepping
            self.step_scale = 2.0 * self.params.native_res_factor
            self.speed_pendown = local_speed_pendown * self.params.speed_lim_xy_hr / 110.0
            self.speed_penup = self.options.speed_penup * self.params.speed_lim_xy_hr / 110.0
            if self.options.const_speed:
                self.speed_pendown = self.speed_pendown * self.params.const_speed_factor_hr
        else:  # i.e., self.options.resolution == 2; Low-resolution ("Normal") mode
            if not self.options.preview:
                res_1, res_2 = ebb_motion.query_enable_motors(self.plot_status.port, False)
                if not (res_1 == 2 and res_2 == 2):  # Do not re-enable if already enabled
                    ebb_motion.sendEnableMotors(self.plot_status.port, 2)  # 8X microstepping
            self.step_scale = self.params.native_res_factor
            # Low-res mode: Allow faster pen-up moves. Keep maximum pen-down speed the same.
            self.speed_penup = self.options.speed_penup * self.params.speed_lim_xy_lr / 110.0
            self.speed_pendown = local_speed_pendown * self.params.speed_lim_xy_lr / 110.0
            if self.options.const_speed:
                self.speed_pendown = self.speed_pendown * self.params.const_speed_factor_lr
        # ebb_serial.command(self.plot_status.port, "CU,3,1\r") # EBB 2.8.1+: Enable data-low LED

    def query_ebb_voltage(self):
        """Check that power supply is detected."""
        serial_utils.query_voltage(self.options, self.params, self.plot_status, self.warnings)

    def get_doc_props(self):
        """
        Get the document's height and width attributes from the <svg> tag. Use a default value in
        case the property is not present or is expressed in units of percentages.
        """

        self.svg_height = plot_utils.getLengthInches(self, 'height')
        self.svg_width = plot_utils.getLengthInches(self, 'width')

        width_string = self.svg.get('width')
        if width_string:
            _value, units = plot_utils.parseLengthWithUnits(width_string)
            self.doc_units = units
        if self.svg_height is None or self.svg_width is None:
            return False
        if self.options.no_rotate:  # Override regular auto_rotate option
            self.options.auto_rotate = False
        if self.options.auto_rotate and (self.svg_height > self.svg_width):
            self.rotate_page = True
        return True

    def get_output(self):
        """Return serialized copy of svg document output"""
        result = etree.tostring(self.document)
        return result.decode('utf-8')

    def disconnect(self):
        """End serial session; disconnect from AxiDraw"""
        if self.plot_status.port:
            ebb_serial.closePort(self.plot_status.port)
        self.plot_status.port = None
        self.connected = False  # Python interactive API variable


class SecondaryLoggingHandler(logging.Handler):
    """To be used for logging to AxiDraw.text_out and AxiDraw.error_out."""

    def __init__(self, axidraw, log_name, level=logging.NOTSET):
        super().__init__(level=level)

        log = getattr(axidraw, log_name) if hasattr(axidraw, log_name) else ''
        setattr(axidraw, log_name, log)

        self.axidraw = axidraw
        self.log_name = log_name

        self.setFormatter(logging.Formatter())  # pass message through unchanged

    def emit(self, record):
        assert hasattr(self.axidraw, self.log_name)
        new_log = getattr(self.axidraw, self.log_name) + '\n' + self.format(record)
        setattr(self.axidraw, self.log_name, new_log)


class SecondaryErrorHandler(SecondaryLoggingHandler):
    """Handle logging for "secondary" machines, plotting alongside primary."""

    def __init__(self, axidraw):
        super().__init__(axidraw, 'error_out', logging.ERROR)


class SecondaryNonErrorHandler(SecondaryLoggingHandler):
    class ExceptErrorsFilter(logging.Filter):
        def filter(self, record):
            return record.levelno < logging.ERROR

    def __init__(self, axidraw):
        super().__init__(axidraw, 'text_out')
        self.addFilter(self.ExceptErrorsFilter())


if __name__ == '__main__':
    logging.basicConfig()
    e = AxiDraw()
    exit_status.run(e.affect)
